import numpy as np
import pandas as pd

from . import accessor


@accessor.register_obscollection_accessor("geo")
class GeoAccessor:
    def __init__(self, oc_obj):
        self._obj = oc_obj

    def get_bounding_box(self, xcol='x', ycol='y', buffer=0):
        """returns the bounding box of all observations.

        Parameters
        ----------
        xcol : str, optional
            column name with x values
        ycol : str, optional
            column name with y values
        buffer : int or float, optional
            add a buffer around the bouding box from the observations

        Returns
        -------
        xmin, ymin, xmax, ymax
            coordinates of bouding box
        """

        xmin = self._obj[xcol].min() - buffer
        xmax = self._obj[xcol].max() + buffer
        ymin = self._obj[ycol].min() - buffer
        ymax = self._obj[ycol].max() + buffer

        return (xmin, ymin, xmax, ymax)

    def get_extent(self, xcol='x', ycol='y', buffer=0):
        """returns the extent of all observations.

        Parameters
        ----------
        xcol : str, optional
            column name with x values
        ycol : str, optional
            column name with y values
        buffer : int or float, optional
            add a buffer around the bouding box from the observations

        Returns
        -------
        xmin, xmax, ymin, ymax
            coordinates of bouding box
        """

        xmin = self._obj[xcol].min() - buffer
        xmax = self._obj[xcol].max() + buffer
        ymin = self._obj[ycol].min() - buffer
        ymax = self._obj[ycol].max() + buffer

        return (xmin, xmax, ymin, ymax)

    def set_lat_lon(self, in_epsg='epsg:28992', out_epsg='epsg:4326',
                    add_to_meta=True, verbose=False):
        """create columns with lat and lon values of the observation points.

        Parameters
        ----------
        in_epsg : str, optional
            epsg code of current x and y attributes, default (RD new)
        out_epsg : str, optional
            epsg code of desired output, default lat/lon
        add_to_meta : bool, optional
            if True the lat and lon values are added to the observation meta
            dictionary. The default is True.
        verbose : boolean, optional
            Print additional information to the screen (default is False).

        Returns
        -------
        None.
        """

        df_lat_lon = self._obj.geo.get_lat_lon(in_epsg, out_epsg)
        for iname in df_lat_lon.index:
            self._obj._set_metadata_value(iname, 'lat',
                                          df_lat_lon.loc[iname,
                                                         'lat'], add_to_meta,
                                          verbose)
            self._obj._set_metadata_value(iname, 'lon',
                                          df_lat_lon.loc[iname,
                                                         'lon'], add_to_meta,
                                          verbose)

    def get_lat_lon(self, in_epsg='epsg:28992', out_epsg='epsg:4326'):
        """get lattitude and longitude from x and y attributes.

        Parameters
        ----------
        in_epsg : str, optional
            epsg code of current x and y attributes, default (RD new)
        out_epsg : str, optional
            epsg code of desired output, default lat/lon

        Returns
        -------
        pandas.DataFrame
            with columns 'lat' and 'lon'
        """

        df_lat_lon = pd.DataFrame(
            index=self._obj.index, columns=['lat', 'lon'])
        for iname in self._obj.index:
            o = self._obj.loc[iname, 'obs']
            df_lat_lon.loc[iname, ['lat', 'lon']
                           ] = o.geo.get_lat_lon(in_epsg, out_epsg)

        return df_lat_lon

    def get_nearest_point(self, obs_collection2=None, gdf2=None,
                          xcol_obs1='x', ycol_obs1='y',
                          xcol_obs2='x', ycol_obs2='y', verbose=False):
        """get nearest point of another obs collection for each point in the
        current obs collection.

        Parameters
        ----------
        obs_collection2 : ObsCollection, optional
            collection of observations of which the nearest point is found
        gdf2 : GeoDataFrame, optional
            dataframe to look for nearest observation point
        xcol_obs1 : str, optional
            x column in self._obj used to get geometry
        ycol_obs1 : str, optional
            y column in self._obj used to get geometry
        xcol_obs2 : str, optional
            x column in obs_collection2 used to get geometry
        ycol_obs2 : str, optional
            y column in self._obj used to get geometry
        verbose : boolean, optional
            Print additional information to the screen (default is False).

        Returns
        -------
        pandas.DataFrame
            with columns 'nearest point' and 'distance nearest point'
        """

        from shapely.ops import nearest_points

        gdf1 = self._obj.to_gdf(xcol=xcol_obs1, ycol=ycol_obs1)
        if obs_collection2 is not None:
            gdf2 = obs_collection2.to_gdf(xcol=xcol_obs2, ycol=ycol_obs2)
        elif gdf2 is None:
            raise ValueError('obs_collecction2 or gdf2 should be defined')

        pts_gdf2 = gdf2.geometry.unary_union

        def nearest_point(point_gdf1, pts=pts_gdf2):
            # find the nearest point and return the corresponding Place value
            nearest_point_gdf2 = nearest_points(point_gdf1, pts_gdf2)[1]
            nearest = gdf2[gdf2.geometry == nearest_point_gdf2]
            return nearest.index[0]

        def distance_nearest_point(point_gdf1, pts=pts_gdf2):
            # find the nearest point and return the corresponding Place value
            nearest_point_gdf2 = nearest_points(point_gdf1, pts_gdf2)[1]
            distance = point_gdf1.distance(nearest_point_gdf2)
            return distance

        gdf1['nearest point'] = gdf1.apply(
            lambda row: nearest_point(row.geometry), axis=1)
        gdf1['distance nearest point'] = gdf1.apply(
            lambda row: distance_nearest_point(row.geometry), axis=1)

        return gdf1[['nearest point', 'distance nearest point']]

    def get_nearest_polygon(self, gdf=None,
                            xcol_obs='x', ycol_obs='y',
                            verbose=False):
        """get nearest polygon for each point in the obs collection.

        Parameters
        ----------
        gdf : GeoDataFrame
            dataframe with polygon features
        xcol_obs : str, optional
            x column in self._obj used to get geometry
        ycol_obs : str, optional
            y column in self._obj used to get geometry
        verbose : boolean, optional
            Print additional information to the screen (default is False).

        Returns
        -------
        pandas.DataFrame
            with columns 'nearest polygon' and 'distance nearest polygon'
        """

        gdf_obs = self._obj.to_gdf(xcol=xcol_obs, ycol=ycol_obs)

        for i, point in gdf_obs.geometry.items():
            distances = [point.distance(pol) for pol in gdf.geometry.values]
            if (np.array(distances) == np.min(distances)).sum() > 1:
                raise ValueError('multiple polygons are nearest')
            gdf_obs.loc[i, 'nearest polygon'] = gdf.iloc[np.argmin(
                distances)].name
            gdf_obs.loc[i, 'distance nearest polygon'] = np.min(distances)

        return gdf_obs[['nearest polygon', 'distance nearest polygon']]

    def get_distance_to_point(self, point, xcol='x', ycol='y'):
        """get distance of every observation to a point.

        Parameters
        ----------
        point : shapely.geometry.point.Point
            point geometry
        xcol : str, optional
            x column in self._obj used to get x coordinates
        ycol : str, optional
            y column in self._obj used to get y coordinates

        Returns
        -------
        pd.Series
            distance to the point for every observation in self._obj
        """

        gdf = self._obj[[xcol, ycol]].to_gdf(xcol=xcol, ycol=ycol)

        return gdf.distance(point)

    def within_extent(self, extent, inplace=False):
        """Slice ObsCollection by extent.

        Parameters
        ----------
        extent : tuple
            format (xmin, xmax, ymin, ymax), default dis.sr.get_extent() format

        Returns
        -------
        new_oc : obs_collection.ObsCollection
            ObsCollection with observations within extent
        """

        new_oc = self._obj[(self._obj.x > extent[0]) & (self._obj.x < extent[1])
                           & (self._obj.y > extent[2]) & (self._obj.y < extent[3])]
        if inplace:
            self._obj._update_inplace(new_oc)
        else:
            return new_oc

    def within_polygon(self, gdf=None, shapefile=None, inplace=False,
                       **kwargs):
        """Slice ObsCollection by checking if points are within a shapefile.

        Parameters
        ----------
        gdf : GeoDataFrame, optional
            geodataframe containing a single polygon
        shapefile : str, optional
            Not yet implemented
        inplace : bool, default False
            Modify the ObsCollection in place (do not create a new object).
        **kwargs :
            kwargs will be passed to the self._obj.to_gdf() method

        Returns
        -------
        new_oc : obs_collection.ObsCollection
            ObsCollection with observations within polygon
        """

        if gdf is not None:
            if gdf.shape[0] != 1:
                raise NotImplementedError(
                    'cannot handle zero or multiple polygons')
            gdf_oc = self._obj.to_gdf(**kwargs)
            new_oc = self._obj.loc[gdf_oc.within(gdf.geometry.values[0])]
        elif shapefile is not None:
            raise NotImplementedError('shapefiles are not yet implemented')
        else:
            raise ValueError('shapefile or gdf must be specified')

        if inplace:
            self._obj._update_inplace(new_oc)
        else:
            return new_oc


@accessor.register_obs_accessor("geo")
class GeoAccessorObs:
    def __init__(self, obs):
        self._obj = obs

    def get_lat_lon(self, in_epsg='epsg:28992', out_epsg='epsg:4326'):
        """get lattitude and longitude from x and y attributes.

        Parameters
        ----------
        in_epsg : str, optional
            epsg code of current x and y attributes, default (RD new)
        out_epsg : str, optional
            epsg code of desired output, default lat/lon

        Returns
        -------
        lon, lat : longitude and lattitude of x, y coordinates
        """

        from pyproj import Proj, transform

        inProj = Proj(in_epsg)
        outProj = Proj(out_epsg)

        if np.isnan(self._obj.x) or np.isnan(self._obj.y):
            lat, lon = np.nan, np.nan
        else:
            lat, lon = transform(inProj, outProj, self._obj.x, self._obj.y)

        return lat, lon



import datetime as dt
import os
import re
import tempfile
import warnings
from timeit import default_timer

import numpy as np
import pandas as pd
import requests
import zeep
from requests.exceptions import HTTPError
from zeep import Plugin
from zeep.plugins import HistoryPlugin
from zeep.wsa import WsAddressingPlugin

from ..util import unzip_file


# %% DINO groundwater CSV methods
def _read_dino_groundwater_header(f):
    line = f.readline()
    header = dict()
    while line not in ['\n', '', '\r\n']:
        if "," in line:  # for csv from dinoloket
            propval = line.split(',')
            prop = propval[0]
            prop = prop.replace(':', '')
            prop = prop.strip()
            val = propval[1]
            if propval[2] != '':
                val = val + ' ' + propval[2].replace(':', '') + ' ' + \
                    propval[3]
            header[prop] = val
        else:  # for artdiver dino-csv
            propval = line.split(":")
            prop = propval[0]
            val = ":".join(propval[1:])
            header[prop] = val
        line = f.readline()

    return line, header


def _read_empty(f, line):
    while (line == '\n') or (line == '\r\n'):
        line = f.readline()
    return line


def _read_dino_groundwater_referencelvl(f, line):
    ref = {}
    while line not in ['\n', '', '\r\n']:
        propval = line.split(',')
        prop = propval[0]
        prop = prop.replace(':', '')
        prop = prop.strip()
        if len(propval) > 1:
            val = propval[1]
            ref[prop] = val
        line = f.readline()
    return line, ref


def _read_dino_groundwater_metadata(f, line):

    _translate_dic_float = {'x-coordinaat': 'x',
                            'y-coordinaat': 'y',
                            'filternummer': 'filternr'}
    _translate_dic_div_100 = {'meetpunt (cm t.o.v. nap)': 'meetpunt',
                              'bovenkant filter (cm t.o.v. nap)': 'bovenkant_filter',
                              'onderkant filter (cm t.o.v. nap)': 'onderkant_filter',
                              'maaiveld (cm t.o.v. nap)': 'maaiveld'}
    metalist = list()
    line = line.strip()
    properties = line.split(',')
    line = f.readline()

    while line not in ['\n', '', '\r\n']:
        meta = dict()
        line = line.strip()
        values = line.split(',')
        for i in range(0, len(values)):
            meta[properties[i].lower()] = values[i]
        metalist.append(meta)
        line = f.readline()

    meta_ts = {}
    if metalist:
        # add time dependent metadata to meta_ts
        for i, meta in enumerate(metalist):
            meta_tsi = {}
            start_date = pd.to_datetime(meta.pop('startdatum'), dayfirst=True)
            end_date = pd.to_datetime(meta.pop('einddatum'), dayfirst=True)
            meta_tsi['locatie'] = meta['locatie']
            for key in _translate_dic_float.keys():
                if meta[key] == '':
                    meta_tsi[_translate_dic_float[key]] = np.nan
                else:
                    meta_tsi[_translate_dic_float[key]] = float(meta[key])

            for key in _translate_dic_div_100.keys():
                if meta[key] == '':
                    meta_tsi[_translate_dic_div_100[key]] = np.nan
                else:
                    meta_tsi[_translate_dic_div_100[key]
                             ] = float(meta[key]) / 100.
            if i == 0:
                for key in meta_tsi.keys():
                    meta_ts[key] = pd.Series(name=key,
                                             dtype=type(meta_tsi[key]))

            for key in meta_tsi.keys():
                meta_ts[key].loc[start_date] = meta_tsi[key]

        # remove series with non time variant metadata from meta_ts
        ts_keys = ['locatie'] + list(_translate_dic_float.values()) + \
            list(_translate_dic_div_100.values())
        for key in ts_keys:
            unique_values = meta_ts[key].unique()
            if len(unique_values) == 1:
                meta_ts.pop(key)

        obs_att = meta_tsi.copy()
        obs_att["name"] = f'{obs_att["locatie"]}-{int(obs_att["filternr"]):03d}'
        obs_att["metadata_available"] = True
    else:
        # de metadata is leeg
        obs_att = {}
        obs_att["locatie"] = ''
        obs_att["filternr"] = np.nan
        obs_att["name"] = 'unknown'
        obs_att["x"] = np.nan
        obs_att["y"] = np.nan
        obs_att["meetpunt"] = np.nan
        obs_att["maaiveld"] = np.nan
        obs_att["bovenkant_filter"] = np.nan
        obs_att["onderkant_filter"] = np.nan
        obs_att["metadata_available"] = False

    return line, obs_att, meta_ts


def _read_dino_groundwater_measurements(f, line):
    line = line.strip()
    titel = line.split(',')
    while '' in titel:
        titel.remove('')

    if line != '':
        # Validate if titles are valid names
        validator = np.lib._iotools.NameValidator()
        titel = [s.lower() for s in validator(titel)]
        usecols = range(0, len(titel))

        try:
            measurements = pd.read_csv(f, header=None, names=titel,
                                       parse_dates=['peildatum'],
                                       index_col='peildatum',
                                       dayfirst=True,
                                       usecols=usecols)
        except pd.errors.ParserError:
            # for now the workflow is to remove the files that cannot be read
            # manually.
            measurements = None
    else:
        measurements = None

    return line, measurements


def read_dino_groundwater_quality_txt(fname, verbose=False):
    """Read dino groundwater quality (grondwatersamenstelling) from a dinoloket
    txt file.

    Notes
    -----
    this function has not been tested thoroughly

    Parameters
    ----------
    fname : str
        path to txt file
    verbose : boolean, optional
        print additional information to the screen (default is False).

    Returns
    -------
    measurements : pd.DataFrame
    meta : dict
        dictionary with metadata
    """

    if verbose:
        print('reading -> {}'.format(os.path.split(fname)[-1]))
    with open(fname, 'r') as f:

        # LOCATIE gegevens
        line = f.readline().rstrip('\n')
        assert line == 'LOCATIE gegevens'

        strt_locatie = f.tell()
        # determine the number of rows
        nrows = -1  # the header does not count
        line = f.readline()
        while line not in ['\n', '']:
            nrows += 1
            line = f.readline()
        eind_locatie = f.tell()
        # go back to where we were before
        f.seek(strt_locatie)
        # read the location-information
        locatie = pd.read_csv(f, sep='\t', nrows=nrows)
        # there is always only one location (change if this is not the case)
        assert nrows == 1

        locatie = locatie.squeeze()

        # KWALITEIT gegevens VLOEIBAAR
        f.seek(eind_locatie)
        line = f.readline().rstrip('\n')
        assert line == 'KWALITEIT gegevens VLOEIBAAR'

        measurements = pd.read_csv(f, sep='\t', parse_dates=['Monster datum', 'Analyse datum'],
                                   dayfirst=True, index_col='Monster datum')

    meta = {}
    meta['filename'] = fname
    meta['locatie'] = locatie['NITG-nr']
    meta['name'] = locatie['NITG-nr']
    meta['x'] = locatie['X-coord']
    meta['y'] = locatie['Y-coord']
    try:
        meta['maaiveld'] = locatie['Maaiveldhoogte (m tov NAP)']
    except KeyError:
        meta['maaiveld'] = np.nan

    return measurements, meta


def read_dino_groundwater_csv(fname, to_mnap=True,
                              read_series=True, verbose=False,):
    """Read dino groundwater quantity data from a dinoloket csv file.

    Parameters
    ----------
    fname : str
        path to csv file
    to_mnap : boolean, optional
        if True a column with 'stand_m_tov_nap' is added to the dataframe
    read_series : boolean, optional
        if False only metadata is read, default is True
    verbose : boolean, optional
        print additional information to the screen (default is False).

    Returns
    -------
    measurements : pd.DataFrame
    meta : dict
        dictionary with metadata
    """
    if verbose:
        print('reading -> {}'.format(os.path.split(fname)[-1]))

    with open(fname, 'r') as f:
        # read header
        line, header = _read_dino_groundwater_header(f)
        line = _read_empty(f, line)
        if verbose and (header == dict()):
            print('could not read header -> {}'.format(fname))

        # read reference level
        line, ref = _read_dino_groundwater_referencelvl(f, line)
        line = _read_empty(f, line)
        if verbose and (ref == dict()):
            print('could not read reference level -> {}'.format(fname))

        # read metadata
        line, meta, meta_ts = _read_dino_groundwater_metadata(f, line)
        line = _read_empty(f, line)
        if verbose and (meta['metadata_available'] == False):
            print('could not read metadata -> {}'.format(fname))
        meta['filename'] = fname

        # read measurements
        if read_series:
            line, measurements = _read_dino_groundwater_measurements(f, line)
            if verbose and (measurements is None):
                print('could not read measurements -> {}'.format(fname))
            elif verbose and measurements[~measurements.stand_cm_tov_nap.isna()].empty:
                print('no NAP measurements available -> {}'.format(fname))
            if to_mnap and measurements is not None:
                measurements['stand_m_tov_nap'] = measurements['stand_cm_tov_nap'] / 100.

            # add time variant metadata to measurements
            for s in meta_ts.values():
                measurements = measurements.join(s, how='outer')
                measurements.loc[:,
                                 s.name] = measurements.loc[:, s.name].ffill()

        else:
            measurements = None

    return measurements, meta


# %% DINO CSV as exported by ArtDiver methods

def _read_artdino_groundwater_metadata(f, line):
    metalist = list()
    line = line.strip()
    properties = line.split(',')
    line = f.readline()
    while line not in ['\n', '', '\r\n']:
        meta = dict()
        line = line.strip()
        values = line.split(',')
        for i in range(0, len(values)):
            meta[properties[i].lower()] = values[i]
        metalist.append(meta)
        line = f.readline()

    meta = {}
    if metalist:
        meta["locatie"] = metalist[-1]['locatie']
        meta["filternr"] = int(float(metalist[-1]['filternummer']))
        meta["name"] = '-'.join([meta["locatie"],
                                 metalist[-1]['filternummer']])
        meta["x"] = float(metalist[-1]['x-coordinaat'])
        meta["y"] = float(metalist[-1]['y-coordinaat'])
        meetpunt = metalist[-1]['meetpunt nap']
        if meetpunt == '':
            meta["meetpunt"] = np.nan
        else:
            meta["meetpunt"] = float(meetpunt) / 100.
        maaiveld = metalist[-1]['maaiveld nap']
        if maaiveld == '':
            meta["maaiveld"] = np.nan
        else:
            meta["maaiveld"] = float(maaiveld) / 100
        bovenkant_filter = metalist[-1]['bovenkant filter']
        if bovenkant_filter == '':
            meta["bovenkant_filter"] = np.nan
        else:
            meta["bovenkant_filter"] = float(bovenkant_filter) / 100
        onderkant_filter = metalist[-1][
            'onderkant filter']
        if onderkant_filter == '':
            meta["onderkant_filter"] = np.nan
        else:
            meta["onderkant_filter"] = float(onderkant_filter) / 100
        meta["metadata_available"] = True
    else:
        # de metadata is leeg
        meta["locatie"] = ''
        meta["filternr"] = np.nan
        meta["name"] = 'unknown'
        meta["x"] = np.nan
        meta["y"] = np.nan
        meta["meetpunt"] = np.nan
        meta["maaiveld"] = np.nan
        meta["bovenkant_filter"] = np.nan
        meta["onderkant_filter"] = np.nan
        meta["metadata_available"] = False

    return line, meta


def _read_artdino_groundwater_measurements(f, line):
    line = line.strip()
    titel = line.split(',')
    while '' in titel:
        titel.remove('')

    if line != '':
        # Validate if titles are valid names
        validator = np.lib._iotools.NameValidator()
        titel = [s.lower() for s in validator(titel)]
        usecols = range(0, len(titel))

        try:
            measurements = pd.read_csv(f, header=None, names=titel,
                                       parse_dates=['peil_datum_tijd'],
                                       index_col='peil_datum_tijd',
                                       dayfirst=True,
                                       usecols=usecols)
        except pd.errors.ParserError:
            # for now the workflow is to remove the files that cannot be read
            # manually.
            measurements = None
    else:
        measurements = None

    return line, measurements


def read_artdino_groundwater_csv(fname, to_mnap=True,
                                 read_series=True, verbose=False,):
    """Read dino groundwater quantity data from a dinoloket csv file.

    Parameters
    ----------
    fname : str
        path to csv file
    to_mnap : boolean, optional
        if True a column with 'stand_m_tov_nap' is added to the dataframe
    read_series : boolean, optional
        if False only metadata is read, default is True
    verbose : boolean, optional
        print additional information to the screen (default is False).

    Returns
    -------
    measurements : pd.DataFrame
    meta : dict
        dictionary with metadata
    """
    if verbose:
        print('reading -> {}'.format(os.path.split(fname)[-1]))

    with open(fname, 'r') as f:
        # read header
        line, header = _read_dino_groundwater_header(f)
        line = _read_empty(f, line)
        if verbose and (header == dict()):
            print('could not read header -> {}'.format(fname))

        # read metadata
        line, meta = _read_artdino_groundwater_metadata(f, line)
        line = _read_empty(f, line)
        if verbose and (meta['metadata_available'] == False):
            print('could not read metadata -> {}'.format(fname))
        meta['filename'] = fname

        # read measurements
        if read_series:
            line, measurements = _read_artdino_groundwater_measurements(
                f, line)
            if verbose and (measurements is None):
                print('could not read measurements -> {}'.format(fname))
            elif verbose and measurements[~measurements["stand_cm_nap"].isna()].empty:
                print('no NAP measurements available -> {}'.format(fname))
            if to_mnap and measurements is not None:
                measurements['stand_m_tov_nap'] = measurements['stand_cm_nap'] / 100.
        else:
            measurements = None

    return measurements, meta


def read_artdino_dir(dirname, ObsClass=None,
                     subdir='csv', suffix='.csv',
                     unpackdir=None, force_unpack=False, preserve_datetime=False,
                     verbose=False, keep_all_obs=True, **kwargs
                     ):
    """Read Dino directory with point observations.

    to do:
    - Evt. nog verbeteren door meteen Dataframe te vullen op het moment dat een observatie
    wordt ingelezen. Nu wordt eerst alles ingelezen in een lijst en daar een dataframe van gemaakt.
    - aparte unzip functie maken en toch de juiste tijdelijke directory krijgen.

    Parameters
    ----------
    dirname : str
        directory name, can be a .zip file or the parent directory of subdir
    ObsClass : type
        class of the observations, e.g. GroundwaterObs or WaterlvlObs
    subdir : str
        subdirectory of dirname with data files
    suffix : str
        suffix of files in subdir that will be read
    unpackdir : str
        destination directory of the unzipped file
    force_unpack : boolean, optional
        force unpack if dst already exists
    preserve_datetime : boolean, optional
        use date of the zipfile for the destination file
    verbose : boolean, optional
        Print additional information to the screen (default is False)
    keep_all_obs : boolean, optional
        add all observation points to the collection, even without data or
        metadata
    **kwargs: dict, optional
        Extra arguments are passed to ObsClass.from_dino_file()

    Returns
    -------
    obs_df : pd.DataFrame
        collection of multiple point observations
    """

    # unzip dir
    if dirname.endswith('.zip'):
        zipf = dirname
        if unpackdir is None:
            dirname = tempfile.TemporaryDirectory().name
        else:
            dirname = unpackdir
        unzip_file(zipf, dirname, force=force_unpack,
                   preserve_datetime=preserve_datetime)

    # read filenames
    files = os.listdir(os.path.join(dirname, subdir))
    if suffix:
        files = [file for file in files if file.endswith(suffix)]

    if not files:
        raise FileNotFoundError('no files were found in {} that end with {}'.format(
            os.path.join(dirname, subdir), suffix))

    # read individual files
    obs_list = []
    for _, file in enumerate(files):
        fname = os.path.join(dirname, subdir, file)
        obs = ObsClass.from_artdino_file(
            fname=fname, verbose=verbose, **kwargs)
        if obs.metadata_available and (not obs.empty):
            obs_list.append(obs)
        elif keep_all_obs:
            obs_list.append(obs)
        else:
            if verbose:
                print('not added to collection -> {}'.format(fname))

    return obs_list


# %% DINO download methods

class RemoveWSA(Plugin):
    """Helper class to remove wsa tags from header in zeep XML post.

    As described by:
    https://github.com/mvantellingen/python-zeep/issues/330#issuecomment-321781637
    """

    def egress(self, envelope, http_headers, operation, binding_options):
        envelope.find(
            '{http://schemas.xmlsoap.org/soap/envelope/}Header').clear()
        return envelope, http_headers


class DinoREST:
    """This object is a first draft for retrieving data from the Dinoloket REST
    API.

    Currently only tested for piezometers.
    """

    aliases = {
        "grondwatermonitoring": "lks_gwo_rd",
        "sondering": "lks_gso_rd",
        "boring": "lks_gbo_rd",
        "peilschaal": "lks_owo_rd"
    }

    def __init__(self, query_url=None, gwo_url=None):
        """DinoREST object for getting metadata from dinoloket.

        Parameters
        ----------
        query_url : str, optional
            url for passing queries to, by default None
        gwo_url : str, optional
            url for getting details for dino piezometers, by default None
        """
        if query_url is None:
            self.query_url = (
                "https://www.dinoloket.nl/arcgis/rest/services/dinoloket/"
                "{layer}/MapServer/0/query")
        else:
            self.query_url = query_url

        if gwo_url is None:
            self.gwo_url = (
                "https://www.dinoloket.nl/javascriptmapviewer-web/"
                "rest/gws/gwo/details")
        else:
            self.gwo_url = gwo_url

    def get(self, url, query):
        """GET method.

        Parameters
        ----------
        url : str
            url
        query : dict
            dictionary containing query parameters

        Returns
        -------
        requests.models.Response
            response from dinoloket REST API
        """
        try:
            response = requests.get(url, params=query)
            response.raise_for_status()
        except HTTPError as http_err:
            raise http_err
        except Exception as err:
            raise err
        else:
            return response

    def post(self, url, json_d):
        """POST method.

        Parameters
        ----------
        url : str
            url
        json_d : dict
            dictionary containing payload as JSON

        Returns
        -------
        requests.models.Response
            response from dinoloket REST API
        """
        try:
            response = requests.post(url, json=json_d)
            response.raise_for_status()
        except HTTPError as http_err:
            raise http_err
        except Exception as err:
            raise err
        else:
            return response

    def query_locations_by_extent(self, extent, layer="grondwatermonitoring"):
        """Get unique locations by extent.

        Parameters
        ----------
        extent : list of floats
            xmin, xmax, ymin, ymax
        layer : str
            dino layer to get data from, see DinoREST.aliases for more
            information on which layer codes to use

        Returns
        -------
        response : requests.models.Response
            query response
        """
        xmin, xmax, ymin, ymax = extent

        query = {
            "f": "pjson",
            "outFields": "*",
            "spatialRel": "esriSpatialRelContains",
            "geometryType": "esriGeometryEnvelope",
            "geometry": f"{xmin},{ymin},{xmax},{ymax}",
            "inSR": 28992,
            "outSR": 28992
        }

        url = self.query_url.format(layer=self.aliases[layer])
        response = self.get(url, query)
        return response

    def get_gwo_details(self, locations):
        """Get metadata details for locations.

        Parameters
        ----------
        locations : list of str
            list of dinoId strings

        Returns
        -------
        requests.models.Response
            response from the Dino REST API
        """
        response = self.post(self.gwo_url, locations)
        return response

    def get_gwo_metadata(self, location, filternr, raw_response=False,
                         verbose=False):
        """Get metadata details for locations.

        Parameters
        ----------
        locations : list of str
            list of dinoId strings

        Returns
        -------
        meta dictionary or requests.models.Response
        """
        response = self.get_gwo_details([location])
        # response = self.post(self.gwo_url, [location])

        if raw_response:
            return response
        else:
            data = response.json()
            if data == []:
                meta = {'metadata_available': False}
            else:
                meta = self._parse_json_single_gwo_filter(data[0], location,
                                                          filternr,
                                                          verbose=verbose)
            return meta

    def get_locations_by_extent(self, extent, layer="grondwatermonitoring",
                                raw=False):
        """get locations within an extent.

        Parameters
        ----------
        extent : list of floats
            xmin, xmax, ymin, ymax
        layer : str, optional
            which dino layer to query, by default "grondwatermonitoring",
            see DinoREST.aliases for more information on which layer codes
            are available
        raw : bool, optional
            return raw JSON response, by default False

        Returns
        -------
        pandas.DataFrame
            DataFrame containing names and locations
        """
        r = self.query_locations_by_extent(extent, layer=layer)

        if raw:
            return r.json()
        else:
            return self._parse_json_features(r.json())

    def get_metadata_by_extent(self, extent, layer="grondwatermonitoring",
                               raw=False):
        """Get metadata details for all objects in extent.

        Parameters
        ----------
        extent : list of floats
            xmin, xmax, ymin, ymax
        layer : str, optional
            which dino layer to query, by default "grondwatermonitoring",
            see DinoREST.aliases for more information on which layer codes
            are available
        raw : bool, optional
            return raw json response, by default False

        Returns
        -------
        pandas.DataFrame
            DataFrame containing metadata per object
        """
        # get locations
        r = self.query_locations_by_extent(extent, layer=layer)
        # get features as json
        loc_json = r.json()["features"]
        # loop through names to create list of locations
        nloc = len(loc_json)
        locations = [loc_json[i]["attributes"]["DINO_NR"] for i in range(nloc)]

        if not locations:
            return pd.DataFrame()
        # get details as json
        rdetails = self.get_gwo_details(locations)
        json_details = rdetails.json()

        if not json_details:
            return pd.DataFrame()

        # return raw or parsed as dataframe
        if raw:
            return json_details
        else:
            if layer == "grondwatermonitoring":
                return self._parse_json_gwo_details(json_details)
            else:
                raise NotImplementedError(
                    "Not yet implemented for '{}'!".format(layer))

    @staticmethod
    def _parse_json_features(js, field="features"):
        data = js[field]
        if len(data) == 0:
            return pd.DataFrame()
        df = pd.DataFrame()
        for idx, idat in enumerate(data):
            for name, value in idat["attributes"].items():
                df.loc[idx, name] = value
            x = idat["geometry"]["x"]
            y = idat["geometry"]["y"]
            df.loc[idx, "x"] = x
            df.loc[idx, "y"] = y
        df.set_index("DINO_NR", inplace=True)
        return df

    @staticmethod
    def _parse_json_single_gwo_filter(data, location, filternr,
                                      verbose=False):
        """parse metadata obtained with get_dino_piezometer_metadata.

        Parameters
        ----------
        data : dictionary
            response from dinoloket
        location : str
            location of piezometer, e.g. B57F0077
        filternr : str
            filter number should be a string of 3 characters, e.g. 002

        Returns
        -------
        meta : dictionary
            parsed dictionary
        """
        _translate_dic_location = {'xcoord': 'x',
                                   'ycoord': 'y',
                                   'surfaceElevation': 'maaiveld'}

        _translate_dic_filter = {'bottomHeightNap': 'onderkant_filter',
                                 'topHeightNap': 'bovenkant_filter'}

        meta = {
            "locatie": location,
            "name": '-'.join([location, filternr]),
            "filternr": int(filternr),
            "metadata_available": True
        }

        piezometers = data.pop('piezoMeters')
        levels = data.pop('levels')
        samples = data.pop('samples')

        if piezometers:
            try:
                meta.update(next(item for item in piezometers if
                                 item["piezometerNr"] == filternr))
            except StopIteration:
                if verbose:
                    print('no piezometer metadata in metadata json '
                          f'{location} with filternr {filternr}')
        if levels:
            try:
                meta.update(next(item for item in levels if
                                 item["piezometerNr"] == filternr))
            except StopIteration:
                if verbose:
                    print('no level metadata in metadata json '
                          f'{location} with filternr {filternr}')
        if samples:
            try:
                meta.update(next(item for item in samples if
                                 item["piezometerNr"] == filternr))
            except StopIteration:
                if verbose:
                    print('no sample metadata in metadata json '
                          f'{location} with filternr {filternr}')

        meta.update(data)
        meta.pop('dinoId')
        for key, item in _translate_dic_location.items():
            meta[item] = meta.pop(key)

        if 'clusterList' in meta.keys():
            if isinstance(meta['clusterList'], list):
                meta['clusterList'] = ";".join(meta['clusterList'])

        if 'piezometerNr' not in meta.keys():
            return meta

        meta.pop('piezometerNr')

        for key, item in _translate_dic_filter.items():
            meta[item] = meta.pop(key)
            if meta[item] is None:
                meta[item] = np.nan

        return meta

    def _parse_json_gwo_details(self, json_details, field="levels",
                                verbose=False):
        """convert json response to dataframe.

        TODO:
        - figure out how to deal with this dictionary to convert to a
        dataframe. There are 3 attributes in the JSON that contain lists:
          - levels
          - samples
          - piezoMeters
          I have no idea what the difference between these is... but it makes
          converting the result to a DataFrame annoying. Currently, I choose one
          of these fields and then skip the rest.
        - currently clusterList field is joined with ';' and added as a string
        instead of the list in JSON.

        Parameters
        ----------
        data : JSON
            json response
        field : str, optional
            default field to use to populate the dataframe, by default "levels"

        Returns
        -------
        pandas.DataFrame
            DataFrame containing metadata for all dino piezometers
        """
        dflist = []
        location_list = []
        for i in range(len(json_details)):
            data = json_details[i]
            location = data['dinoId']

            if location in location_list:
                raise ValueError
            else:
                location_list.append(location)

            piezometers = data['piezoMeters']
            levels = data['levels']
            samples = data['samples']

            # check number of filters
            if piezometers is None:
                piezometers = []
            if levels is None:
                levels = []
            if samples is None:
                samples = []

            # merge piezometers, levels and samples
            filter_dic = {}
            for l in [piezometers, levels, samples]:
                for d in l:
                    if 'piezometerNr' in d.keys():
                        if d['piezometerNr'] not in filter_dic.keys():
                            filter_dic[d['piezometerNr']] = d
                        else:
                            filter_dic[d['piezometerNr']].update(d)

            # get metadata for every filter
            for filternr in filter_dic.keys():
                meta = self._parse_json_single_gwo_filter(
                    data.copy(), location, str(filternr).zfill(3),
                    verbose=verbose)
                idf = pd.DataFrame(meta, index=[meta.pop('name')])
                dflist.append(idf)

        df = pd.concat(dflist, axis=0, sort=True)

        return df


class DinoWSDL:
    """Class for querying DINOLoket. Currently available methods are:

        - findMeetreeks: get timeseries for piezometer
        - findWaarnemingGegevens: get some simple all time statistics
          for a piezometer
        - findTechnischeGegevens: get metadata for a piezometer
        - findGrondwaterStatistiek: get stats for 14th and 28th day of
          each month for a given period

    Returns
    -------
    DinoWSDL :
        returns an object that can be used to query the DINO Webservice.
    """

    def __init__(self, wsdl="http://www.dinoservices.nl/gwservices/gws-v11?wsdl"):
        """Initialize DinoWSDL object that can be used to query DINO
        Webservice.

        Parameters
        ----------
        wsdl : str, optional
            wsdl url, by default
            "http://www.dinoservices.nl/gwservices/gws-v11?wsdl"
        """

        raise RuntimeError('Dino WSDL API is down since January 2021, hopefully something new will arrive soon!')

        # Create some plugins, some currently unused but left here as a reminder.
        history = HistoryPlugin()
        wsa = WsAddressingPlugin()
        rwsa = RemoveWSA()

        # Configure the client
        self.client = zeep.Client(
            wsdl=wsdl,
            plugins=[history, wsa, rwsa])

        # Set prefix
        self.client.set_ns_prefix("v11", "http://v11.ws.gws.dino.tno.nl/")

    def findMeetreeks(self, location, filternr, tmin, tmax, unit="NAP",
                      raw_response=False, verbose=False):
        """Get a timeseries for a piezometer. If the piezometer is part of a
        cluster this function returns the combined timeseries of all the
        piezometers in the cluster.

        Parameters
        ----------
        location : str
            location str of the piezometer, i.e. B57F0077
        filternr : str, int, float
            filter number, is converted to str, i.e. 004
        tmin : str or pandas.Timestamp
            start date in format YYYY-MM-DD (will be converted if Timestamp)
        tmax : str or pandas.Timestamp
            end date in format YYYY-MM-DD (will be converted if Timestamp)
        unit : str, optional
            unit in which timeseries is returned, by default "NAP"
        raw_response : bool, optional
            if True, return the zeep parsed XML response, by default
            False which will return a DataFrame
        verbose : bool, optional
            print logging information to console

        Returns
        -------
        df or response
            returns DataFrame or zeep parsed XML response

        Note
        ----
        - method sometimes returns data
        """
        if isinstance(tmin, pd.Timestamp):
            tmin = tmin.strftime('%Y-%m-%d')

        if isinstance(tmax, pd.Timestamp):
            tmax = tmax.strftime('%Y-%m-%d')

        if unit not in ["NAP", "SFL", "MP"]:
            raise ValueError("'unit' must be one of 'NAP', 'SFL', 'MP'!")

        data = {"WELL_NITG_NR": location,
                "WELL_TUBE_NR": filternr,
                "START_DATE": tmin,
                "END_DATE": tmax,
                "UNIT": unit}

        try:
            start = default_timer()
            if verbose:
                print(f"Downloading: {location}-{filternr}... ", end="",
                      flush=True)
            r = self.client.service.findMeetreeks(**data)
        except Exception as e:
            if verbose:
                print(e)
            raise e
        time = default_timer() - start
        if verbose:
            if r:
                print(f"OK! {time:.3f}s", flush=True)
            else:
                print(f"ERROR {r.status_code}!", flush=True)

        if raw_response:
            return r
        else:
            df = self._parse_grondwaterstand(r, f'stand_m_tov_{unit.lower()}')

            return df

    def findTechnischeGegevens(self, location, filter_nr, raw_response=False):
        """Get metadata for a piezometer.

        Parameters
        ----------
        location : str
            location str of the piezometer, i.e. B57F0077
        filter_nr : str, int, float
            filter number, is converted to str, i.e. 004
        raw_response : bool, optional
            if True, return the zeep parsed XML response, by default
            False which will return a DataFrame

        Returns
        -------
        df or response
            returns DataFrame or zeep parsed XML response
        """

        filter_nr = "{0:03g}".format(int(filter_nr))

        data = {"WELL_NITG_NR": location,
                "WELL_TUBE_NR": filter_nr}

        try:
            r = self.client.service.findTechnischeGegevens(**data)
        except Exception as e:
            raise e

        if raw_response:
            return r
        else:
            meta = self._parse_technische_gegevens(r)
            return meta

    @staticmethod
    def _parse_grondwaterstand(r, column_name='stand_m_tov_nap'):
        """Parse the response from findMeetreeks.

        Parameters
        ----------
        r : zeep.objects.GrondwaterStanden
            the response of findMeetreeks()

        Returns
        -------
        pandas.DataFrame
            pandas.DataFrame containing the data
        """
        if isinstance(r, list):
            gws = r[0]

        index = [i["DATE"] for i in gws.LEVELS]
        values = [i["LEVEL"] for i in gws.LEVELS]
        remarks = [i["REMARK"] for i in gws.LEVELS]
        df = pd.DataFrame(index=index, data={column_name: values,
                                             'remarks': remarks})
        # convert from cm to m
        df.loc[:, column_name] = df.loc[:, column_name] / 100

        return df

    @staticmethod
    def _parse_technische_gegevens(response):
        """Parse the response from findTechnischeGegevens.

        Parameters
        ----------
        response : zeep.objects.TechnischeGegevens
            the response of findTechnischeGegevens()

        Returns
        -------
        pandas.DataFrame
            pandas.DataFrame containing the data
        """
        if isinstance(response, list):
            response = response[0]

        locatie = response.WELL_NITG_NR
        filternr = response.WELL_TUBE_NR
        meta = {
            "locatie": locatie,
            "name": '-'.join([locatie, filternr]),
            "filternr": int(filternr)
        }

        if response.SFL_HEIGHT is not None:
            meta["maaiveld"] = response.SFL_HEIGHT / 100.
        else:
            meta["maaiveld"] = np.nan

        if response.BOTTOM_FILTER is not None:
            meta["onderkant_filter"] = response.BOTTOM_FILTER / 100.
        else:
            meta["onderkant_filter"] = np.nan

        if response.FILTER_LENGTH is not None:
            meta["filterlengte"] = response.FILTER_LENGTH / 100.
        else:
            meta["filterlengte"] = np.nan

        meta["bovenkant_filter"] = meta["onderkant_filter"] + meta["filterlengte"]

        return meta


def download_dino_groundwater(location, filternr, tmin, tmax,
                              split_cluster=True,
                              verbose=False, **kwargs):
    """Download measurements and metadata from a dino groundwater observation
    well.

    Parameters
    ----------
    location : str
        location str of the piezometer, i.e. B57F0077
    filternr : int or float
        filter number, is converted to str, i.e. 004
    tmin : str or pandas.Timestamp
        start date in format YYYY-MM-DD (will be converted if Timestamp)
    tmax : str or pandas.Timestamp
        end date in format YYYY-MM-DD (will be converted if Timestamp)
    split_cluster : bool
        if False and the piezometer belongs to a cluster, the combined
        time series of the cluster is used. if True the indvidual time
        series of each piezometer is used. Default is True
    verbose : bool
        print logging inforation to console, default is False
    kwargs : key-word arguments
            these arguments are passed to dino.findMeetreeks functie

    Returns
    -------
    measurements : pd.DataFrame
    meta : dict
        dictionary with metadata
    """
    filternr = "{0:03g}".format(int(filternr))

    # download data from dino
    dino = DinoWSDL()

    # measurements
    measurements = dino.findMeetreeks(location, filternr, tmin, tmax,
                                      verbose=verbose, **kwargs)

    # new metadata method
    dinorest = DinoREST()
    meta = dinorest.get_gwo_metadata(location, filternr, verbose=verbose)

    if ('clusterList' in meta.keys()) and split_cluster and (not measurements.empty):
        if verbose:
            print(f'piezometer {location}-{filternr} is part of a cluster')
            print('slicing time series to avoid using combined cluster series')

        start_time = dt.timedelta(
            seconds=meta['startDate'] / 1000) + dt.datetime(1970, 1, 1, 1)
        end_time = dt.timedelta(
            seconds=meta['endDate'] / 1000) + dt.datetime(1970, 1, 1, 1)
        measurements = measurements[start_time:end_time]

    return measurements, meta


def download_dino_groundwater_bulk(locations, ObsClass=None,
                                   filtersep="-", verbose=False,
                                   stop_on_error=True,
                                   **kwargs):
    """Download DINO groundwater level timeseries based on a list of locations.

    Parameters
    ----------
    locations : list of str
        list of locations including filter numbers (i.e. B00H0001-001)
    ObsClass : Obs, optional
        ObsClass to use for storing timeseries, by default None. It is
        recommended to use GroundwaterObs.
    filtersep : str, optional
        separation character(s) between location code and filter
        number, by default "-"
    verbose : bool, optional
        print logging information to console, by default False
    stop_on_error : bool, optional
        if download fails for some reason, raise the error, otherwise
        continue, by default True

    Returns
    -------
    obs_list : list of ObsClass
        list of ObsClass objects containing dino groundwater timeseries and
        metadata
    """
    obslist = []
    for loc in locations:
        iloc, filternr = loc.split(filtersep)
        try:
            obslist.append(ObsClass.from_dino(
                location=iloc, filternr=filternr, verbose=verbose, **kwargs))
        except Exception as e:
            if stop_on_error:
                raise e
            else:
                if verbose:
                    print(loc, e)
                continue

    return obslist


def get_dino_locations(extent=None, bbox=None, layer='grondwatermonitoring',
                       response_fname=None, split_dx=None):
    """Get a GeoDataFrame with dinoloket-locations from a wfs.

    Parameters
    ----------
    extent : list, tuple or numpy-array (user must specify extent or bbox)
        The extent, in RD-coordinates, for which you want to retreive locations
        [xmin, xmax, ymin, ymax]
    bbox : list, tuple or numpy-array (user must specify extent or bbox)
        The bounding box, in RD-coordinates, for which you want to retreive locations
        [xmin, ymin, xmax, ymax]
    layer : str, optional
        The type of timeseries (grondwatermonitoring or boring)

    Returns
    -------
    df : pandas DataFrame or a tuple of GeoDataFrame

    if kind=None
        a geodataframe with the properties of the locations
    """

    # get extent
    if extent is None and bbox is None:
        raise (Exception('Either bbox or extent needs to be specified'))
    if extent is not None and bbox is not None:
        raise (Exception('bbox and extent cannot both be specified'))
    if bbox is not None:
        extent = [bbox[0], bbox[2], bbox[1], bbox[3]]
        if split_dx is not None:
            gdf = split_extent(get_dino_locations, extent,
                               split_dx, layer=layer)
            return gdf
    if isinstance(extent, np.ndarray):
        extent = extent.tolist()

    # use dino REST api
    if not layer in ['grondwatermonitoring', 'boring']:
        raise(Exception('Unknown layer: {}'.format(layer)))
    dinorest = DinoREST()
    if layer == 'boring':
        df = dinorest.get_locations_by_extent(extent, layer=layer)
    elif layer == 'grondwatermonitoring':
        df = dinorest.get_metadata_by_extent(extent, layer=layer)
    else:
        raise NotImplementedError(
            "Not yet implemented for '{}'!".format(layer))

    # check if there are any measurements
    if df.empty:
        warnings.warn('no observation points were found')
        return df

    return df


def split_extent(method, extent0, dx=1000, dy=None, **kwargs):
    if dy is None:
        dy = dx
    ncol = int(np.ceil((extent0[1] - extent0[0]) / dx))
    nrow = int(np.ceil((extent0[3] - extent0[2]) / dy))
    x = np.linspace(extent0[0], extent0[1], ncol)
    y = np.linspace(extent0[2], extent0[3], nrow)
    for ix in range(len(x) - 1):
        for iy in range(len(y) - 1):
            print('Downloading part {} of {}'.format(
                ix * iy + iy + 1, len(x) * len(y)))
            extent = [x[ix], x[ix + 1], y[iy], y[iy + 1]]
            gdft = method(extent=extent, **kwargs)
            if ix == 0 and iy == 0:
                gdf = gdft
            else:
                gdf = pd.concat((gdf, gdft))


def download_dino_within_extent(extent=None, bbox=None, ObsClass=None,
                                layer='grondwatermonitoring',
                                tmin="1900-01-01", tmax="2040-01-01",
                                zmin=None, zmax=None, unit="NAP",
                                keep_all_obs=True,
                                cache=False,
                                verbose=False):
    """Download DINO data within a certain extent (or a bounding box)

    Parameters
    ----------
    extent : list, tuple or numpy-array (user must specify extent or bbox)
        The extent, in RD-coordinates, for which you want to retreive locations
        [xmin, xmax, ymin, ymax]
    bbox : list, tuple or numpy-array (user must specify extent or bbox)
        The bounding box, in RD-coordinates, for which you want to retreive locations
        [xmin, ymin, xmax, ymax]
    ObsClass : type
        class of the observations, so far only GroundwaterObs is supported
    layer : str
        The layer of timeseries ('grondwatermonitoring' of 'boring')
    tmin : str or pandas.Timestamp
        start date in format YYYY-MM-DD (will be converted if Timestamp)
    tmax : str or pandas.Timestamp
        end date in format YYYY-MM-DD (will be converted if Timestamp)
    zmin : float, optional
        minimum filter height in m NAP for which piezometers are returned
    zmax : float, optional
        maximum filter height in m NAP for which piezometers are returned
    unit : str, optional
        unit in which the timeseries are returned, by default "NAP"
    keep_all_obs : boolean, optional
        add all observation points to the collection, even without data or
        metadata
    cache : boolean, optional
        if True each observation that has been downloaded is cached. Can be
        helpful because the dino server give errors frequently. default is
        False
    verbose : boolean, optional
        print additional information to the screen (default is False).

    Returns
    -------
    obs_df : pd.DataFrame
        collection of multiple point observations
    """

    # read locations
    gdf_loc = get_dino_locations(extent=extent, bbox=bbox, layer=layer)

    if verbose:
        if bbox is None:
            print('\ndownload {layer} data from dino within:\n'
                  f'- extent: {extent}')
        elif extent is None:
            print('\ndownload {layer} data from dino within:\n'
                  f'- bbox: {bbox}')

    if gdf_loc.empty:
        return pd.DataFrame()

    gdf_loc["startDate"] = gdf_loc.startDate.astype('datetime64[ms]')
    gdf_loc["endDate"] = gdf_loc.endDate.astype('datetime64[ms]')

    # slice by properties
    if tmin is not None:
        tmin = pd.to_datetime(tmin)
        mask = gdf_loc.endDate > tmin
        gdf_loc = gdf_loc.loc[mask]
    if tmax is not None:
        tmax = pd.to_datetime(tmax)
        mask = gdf_loc.startDate < tmax
        gdf_loc = gdf_loc.loc[mask]
    if zmin is not None:
        mask = gdf_loc.topHeightNap >= zmin
        gdf_loc = gdf_loc.loc[mask]
    if zmax is not None:
        mask = gdf_loc.bottomHeightNap <= zmax
        gdf_loc = gdf_loc.loc[mask]

    if gdf_loc.empty:
        if verbose:
            print("no data found!")
        return pd.DataFrame()

    # read measurements
    obs_list = []
    for index, loc in gdf_loc.iterrows():
        if tmin is None:
            tmin_t = loc.startDate
        else:
            tmin_t = tmin
        if tmax is None:
            tmax_t = loc.endDate
        else:
            tmax_t = tmax

        if verbose:
            print('reading -> {}'.format(index))

        if cache:
            cache_dir = os.path.join(tempfile.gettempdir(), 'dino')
            if not os.path.isdir(cache_dir):
                os.mkdir(cache_dir)
            fname = os.path.join(
                cache_dir, f'{loc.locatie}-{loc.filternr:03d}' + '.pklz')
            if os.path.isfile(fname):
                o = pd.read_pickle(fname)
            else:
                o = ObsClass.from_dino(location=loc.locatie,
                                       filternr=float(loc.filternr),
                                       tmin=tmin_t,
                                       tmax=tmax_t,
                                       unit=unit)
                o.to_pickle(fname)
        else:
            o = ObsClass.from_dino(location=loc.locatie,
                                   filternr=float(loc.filternr),
                                   tmin=tmin_t,
                                   tmax=tmax_t,
                                   unit=unit)

        if o.metadata_available and (not o.empty):
            obs_list.append(o)
        elif keep_all_obs:
            obs_list.append(o)
        else:
            if verbose:
                print('not added to collection -> {}'.format(o.name))

    # return list of obs
    return obs_list


# %% DINO waterlevel observations


def _read_dino_waterlvl_metadata(f, line):
    '''

    Parameters
    ----------
    f : text wrapper
    line : str
        line with meta dictionary keys
    meta_dic : dict (optional)
        dictionary with metadata

    Returns
    -------
    meta : dict
        dictionary with metadata

    '''
    meta_keys = line.strip().split(',')
    meta_values = f.readline().strip().split(',')
    meta = {}
    for key, value in zip(meta_keys, meta_values):
        key = key.strip()
        if key in ['X-coordinaat', 'Y-coordinaat']:
            if key == 'X-coordinaat':
                meta['x'] = float(value)
            elif key == 'Y-coordinaat':
                meta['y'] = float(value)
        elif key == 'Locatie':
            meta['locatie'] = value
            meta['name'] = value

    return meta


def _read_dino_waterlvl_measurements(f, line):
    '''

    Parameters
    ----------
    f : text wrapper
    line: str
        header of dataframe

    Returns
    -------
    measurements : pd.DataFrame
    '''

    titel = line.strip().split(',')
    while '' in titel:
        titel.remove('')

    validator = np.lib._iotools.NameValidator()
    titel = [i.lower() for i in validator(titel)]
    usecols = range(0, len(titel))

    measurements = pd.read_csv(f, header=None, names=titel,
                               parse_dates=['peildatum'],
                               index_col='peildatum',
                               dayfirst=True,
                               usecols=usecols)

    # measurements['Stand (m t.o.v. NAP)'] = measurements['Stand (cm t.o.v. NAP)'] /100.

    # measurements.set_index('Peildatum', inplace=True)

    return measurements


def read_dino_waterlvl_csv(fname, to_mnap=True, read_series=True, verbose=False):
    """Read dino waterlevel data from a dinoloket csv file.

    Parameters
    ----------
    fname : str
    to_mnap : boolean, optional
        if True a column with 'stand_m_tov_nap' is added to the dataframe
    read_series : boolean, optional
        if False only metadata is read, default is True
    verbose : boolean, optional
        print additional information to the screen (default is False).
    """
    if verbose:
        print('reading -> {}'.format(os.path.split(fname)[-1]))

    p_meta = re.compile(
        'Locatie,Externe aanduiding,X-coordinaat,Y-coordinaat, Startdatum, Einddatum')
    p_data = re.compile(
        r'Locatie,Peildatum,Stand \(cm t.o.v. NAP\),Bijzonderheid')

    with open(fname, 'r') as f:
        line = f.readline()
        while line != '':
            line = f.readline()
            if p_meta.match(line):
                meta = _read_dino_waterlvl_metadata(f, line)
                if meta:
                    meta['metadata_available'] = True
                else:
                    meta['metadata_available'] = False
                meta['filename'] = fname
            elif p_data.match(line):
                if read_series:
                    measurements = _read_dino_waterlvl_measurements(f, line)
                    if to_mnap and measurements is not None:
                        measurements['stand_m_tov_nap'] = measurements['stand_cm_tov_nap'] / 100.
                else:
                    measurements = None

                return measurements, meta


# %% General DINO methods
def read_dino_dir(dirname, ObsClass=None,
                  subdir='Boormonsterprofiel_Geologisch booronderzoek', suffix='.txt',
                  unpackdir=None, force_unpack=False, preserve_datetime=False, verbose=False,
                  keep_all_obs=True, **kwargs
                  ):
    """Read Dino directory with point observations.

    to do:
    - Evt. nog verbeteren door meteen Dataframe te vullen op het moment dat een observatie
    wordt ingelezen. Nu wordt eerst alles ingelezen in een lijst en daar een dataframe van gemaakt.
    - aparte unzip functie maken en toch de juiste tijdelijke directory krijgen.

    Parameters
    ----------
    dirname : str
        directory name, can be a .zip file or the parent directory of subdir
    ObsClass : type
        class of the observations, e.g. GroundwaterObs or WaterlvlObs
    subdir : str
        subdirectory of dirname with data files
    suffix : str
        suffix of files in subdir that will be read
    unpackdir : str
        destination directory of the unzipped file
    force_unpack : boolean, optional
        force unpack if dst already exists
    preserve_datetime : boolean, optional
        use date of the zipfile for the destination file
    verbose : boolean, optional
        Print additional information to the screen (default is False)
    keep_all_obs : boolean, optional
        add all observation points to the collection, even without data or
        metadata
    **kwargs: dict, optional
        Extra arguments are passed to ObsClass.from_dino_file()

    Returns
    -------
    obs_df : pd.DataFrame
        collection of multiple point observations
    """

    # unzip dir
    if dirname.endswith('.zip'):
        zipf = dirname
        if unpackdir is None:
            dirname = tempfile.TemporaryDirectory().name
        else:
            dirname = unpackdir
        unzip_file(zipf, dirname, force=force_unpack,
                   preserve_datetime=preserve_datetime)

    # read filenames
    files = os.listdir(os.path.join(dirname, subdir))
    if suffix:
        files = [file for file in files if file.endswith(suffix)]

    if not files:
        raise FileNotFoundError('no files were found in {} that end with {}'.format(
            os.path.join(dirname, subdir), suffix))

    # read individual files
    obs_list = []
    for _, file in enumerate(files):
        fname = os.path.join(dirname, subdir, file)
        obs = ObsClass.from_dino(fname=fname, verbose=verbose, **kwargs)
        if obs.metadata_available and (not obs.empty):
            obs_list.append(obs)
        elif keep_all_obs:
            obs_list.append(obs)
        else:
            if verbose:
                print('not added to collection -> {}'.format(fname))

    return obs_list



# -*- coding: utf-8 -*-
"""Created on Thu Oct 10 11:01:22 2019.

@author: oebbe
"""

import os

import numpy as np
from hydropandas import observation
from pandas import DataFrame, Series
from scipy.io import loadmat

from ..util import matlab2datetime


def read_file(fname, ObsClass, verbose=False):
    """This method is used to read the file."""

    if verbose:
        print('reading menyanthes file {}'.format(fname))

    if ObsClass == observation.GroundwaterObs:
        _rename_dic = {'xcoord': 'x',
                       'ycoord': 'y',
                       'upfiltlev': 'bovenkant_filter',
                       'lowfiltlev': 'onderkant_filter',
                       'surflev': 'maaiveld',
                       'filtnr': 'filternr',
                       'meetpunt': 'measpointlev'
                       }

        _keys_o = ['name', 'x', 'y', 'locatie', 'filternr',
                   'metadata_available', 'maaiveld', 'meetpunt',
                   'bovenkant_filter', 'onderkant_filter']

    elif ObsClass == observation.WaterlvlObs:
        _rename_dic = {'xcoord': 'x',
                       'ycoord': 'y',
                       'meetpunt': 'measpointlev'
                       }

        _keys_o = ['name', 'x', 'y', 'locatie']

    # Check if file is present
    if not (os.path.isfile(fname)):
        print('Could not find file ', fname)

    mat = loadmat(fname, struct_as_record=False, squeeze_me=True,
                  chars_as_strings=True)

    d_h = read_oseries(mat)

    locations = d_h.keys()
    obs_list = []
    for location in locations:
        if verbose:
            print('reading location -> {}'.format(location))
        metadata = d_h[location]
        metadata['projection'] = 'epsg:28992'
        metadata['metadata_available'] = True

        s = metadata.pop('values')
        df = DataFrame(s, columns=['stand_m_tov_nap'])
        for key in _rename_dic.keys():
            if key in metadata.keys():
                metadata[_rename_dic[key]] = metadata.pop(key)

        meta_o = {k: metadata[k] for k in _keys_o if k in metadata}

        o = ObsClass(df, meta=metadata, **meta_o)
        obs_list.append(o)

    return obs_list


def read_oseries(mat):
    """Read the oseries from a mat file from menyanthes."""
    d_h = {}
    # Check if more then one time series model is present
    if not isinstance(mat['H'], np.ndarray):
        mat['H'] = [mat['H']]

    # Read all the time series models
    for i, H in enumerate(mat['H']):
        data = {}

        for name in H._fieldnames:
            if name != 'values':
                data[name] = getattr(H, name)
            else:
                if H.values.size == 0:
                    # when diver-files are used, values will be empty
                    series = Series()
                else:
                    tindex = map(matlab2datetime, H.values[:, 0])
                    # measurement is used as is
                    series = Series(H.values[:, 1], index=tindex)
                    # round on seconds, to get rid of conversion milliseconds
                    series.index = series.index.round('s')
                data['values'] = series

        # add to self.H
        if not hasattr(H, 'Name') and not hasattr(H, 'name'):
            H.Name = 'H' + str(i)  # Give it the index name
        if hasattr(H, 'name'):
            H.Name = H.name
        if len(H.Name) == 0:
            H.Name = H.tnocode

        d_h[H.Name] = data

    return d_h



import os

import numpy as np

from . import accessor


@accessor.register_obscollection_accessor("plots")
class CollectionPlots:

    def __init__(self, oc_obj):
        """Object containing plotting methods for ObsCollections.

        Parameters
        ----------
        oc : ObsCollection
            ObsCollection instance
        """
        self._obj = oc_obj

    def interactive_plots(self, savedir,
                          tmin=None, tmax=None,
                          per_location=True,
                          verbose=True, **kwargs):
        """Create interactive plots of the observations using bokeh.

        Parameters
        ----------
        savedir : str
            directory used for the folium map and bokeh plots
        tmin : dt.datetime, optional
            start date for timeseries plot
        tmax : dt.datetime, optional
            end date for timeseries plot
        per_location : bool, optional
            if True plot multiple filters on the same location in one figure
        verbose : boolean, optional
            Print additional information to the screen (default is False).
        **kwargs :
            will be passed to the Obs.to_interactive_plot method, options
            include:

            - plot_columns : list of str
            - hoover_names : list of str
            - plot_freq : str
            - plot_legend_name : str
            - ylabel : str
            - color : str
            - add_filter_to_legend : boolean
        """
        _color_cycle = (
            'blue',
            'olive',
            'lime',
            'red',
            'orange',
            'yellow',
            'purple',
            'silver',
            'powderblue',
            'salmon',
            'tan')

        if per_location:
            plot_names = self._obj.groupby('locatie').count().index
        else:
            plot_names = self._obj.index

        for name in plot_names:

            if per_location:
                oc = self._obj.loc[self._obj.locatie ==
                                   name, 'obs'].sort_index()
            else:
                oc = self._obj.loc[[name], 'obs']

            p = None
            for i, o in enumerate(oc.values):
                if i == 10:
                    raise NotImplementedError(
                        'cannot add more than 10 lines to a single plot')
                try:
                    p = o.plots.interactive_plot(savedir=savedir,
                                                 p=p,
                                                 tmin=tmin, tmax=tmax,
                                                 colors=[_color_cycle[i + 1]],
                                                 return_filename=False,
                                                 **kwargs)
                    if verbose:
                        print('created iplot -> {}'.format(o.name))

                except ValueError:
                    if verbose:
                        print('{} has no data between {} and {}'.format(
                            o.name, tmin, tmax))
                    o.iplot_fname = None

    def interactive_map(self, plot_dir, m=None,
                        tiles='OpenStreetMap',
                        fname=None,
                        per_location=True,
                        color='blue',
                        legend_name=None,
                        add_legend=True,
                        map_label='',
                        map_label_size=20,
                        col_name_lat='lat',
                        col_name_lon='lon',
                        zoom_start=13,
                        create_interactive_plots=True,
                        verbose=False,
                        **kwargs):
        """Create an interactive map with interactive plots using folium and
        bokeh.

        Notes
        -----
        Some notes on this method:

        - if you want to have multiple obs collections on one folium map, only
          the last one should have add_legend = True to create a correct legend
        - the color of the observation point on the map is now the same color
          as the line of the observation measurements. Also a built-in color
          cycle is used for different measurements on the same location.

        Parameters
        ----------
        plot_dir : str
            directory used for the folium map and bokeh plots
        m : folium.Map, str, optional
            current map to add observations too, if None a new map is created
        tiles : str, optional
            background tiles, default is openstreetmap
        fname : str, optional
            name of the folium map
        per_location : bool, optional
            if True plot multiple filters on the same location in one figure
        color : str, optional
            color of the observation points on the map
        legend_name : str, optional
            the name of the observation points shown in the map legend
        add_legend : boolean, optional
            add a legend to a plot
        map_label : str, optional
            add a label to the obs locations on the map, this label is
            picked from the meta attribute of the obs points.
        map_label_size : int, optional
            label size of the map_label in pt.
        col_name_lat : str, optional
            name of the column in the obs_collection dic with the lat values
            of the observation points
        col_name_lon : str, optional
            see col_name_lat
        zoom_start : int, optional
            start zoom level of the folium ma
        create_interactive_plots : boolean, optional
            if True interactive plots will be created, if False the iplot_fname
            attribute of the observations is used.
        verbose : boolean, optional
            Print additional information to the screen (default is False).
        **kwargs :
            will be passed to the to_interactive_plots method options are:

            - plot_columns : list of str
            - hoover_names : list of str
            - plot_legend_name : str
            - ylabel : str
            - add_filter_to_legend : boolean
            - plot_freq : str
            - tmin : dt.datetime
            - tmax : dt.datetime

        Returns
        -------
        m : folium.Map
            the folium map
        """

        import branca
        import folium
        from folium.features import DivIcon

        # create interactive bokeh plots
        if create_interactive_plots:
            self._obj.plots.interactive_plots(savedir=plot_dir,
                                              verbose=verbose,
                                              per_location=per_location,
                                              **kwargs)

        # determine start location of map
        northing = np.mean(
            (self._obj[col_name_lat].min(), self._obj[col_name_lat].max()))
        easting = np.mean(
            (self._obj[col_name_lon].min(), self._obj[col_name_lon].max()))

        # create map if no map is given
        if m is None:
            m = folium.Map([northing, easting], zoom_start=zoom_start)

        # add the point observations with plots to the map
        group_name = '<span style=\\"color: {};\\">{}</span>'.format(
            color, legend_name)
        group = folium.FeatureGroup(name=group_name)

        if per_location:
            plot_names = self._obj.groupby('locatie').count().index
        else:
            plot_names = self._obj.index

        for name in plot_names:
            if per_location:
                oc = self._obj.loc[self._obj.locatie ==
                                   name, 'obs'].sort_index()
                o = oc.iloc[-1]
                name = o.name
            else:
                o = self._obj.loc[name, 'obs']

            if o.iplot_fname is not None:
                with open(o.iplot_fname, 'r') as f:
                    bokeh_html = f.read()

                iframe = branca.element.IFrame(
                    html=bokeh_html, width=620, height=420)
                popup = folium.Popup(iframe, max_width=620)

                folium.CircleMarker([self._obj.loc[o.name, col_name_lat],
                                     self._obj.loc[o.name, col_name_lon]],
                                    icon=folium.Icon(icon='signal'), fill=True,
                                    color=color,
                                    popup=popup).add_to(group)

                if map_label != '':
                    folium.map.Marker(
                        [self._obj.loc[name, col_name_lat], self._obj.loc[name, col_name_lon]], icon=DivIcon(
                            icon_size=(
                                150, 36), icon_anchor=(
                                0, 0), html='<div style="font-size: %ipt">%s</div>' %
                            (map_label_size, o.meta[map_label]))).add_to(group)
            elif verbose:
                print('no iplot available for {}'.format(o.name))

        group.add_to(m)

        # add legend
        if add_legend:
            folium.map.LayerControl('topright', collapsed=False).add_to(m)

        # save map
        #filename and path
        if fname is not None:
            if not fname.endswith('.html'):
                fname = fname + '.html'
            if not os.path.exists(plot_dir):
                os.mkdir(plot_dir)
            m.save(os.path.join(plot_dir, fname))

        return m


@accessor.register_obs_accessor("plots")
class ObsPlots:
    def __init__(self, obs):
        self._obj = obs

    def interactive_plot(self,
                         savedir=None,
                         plot_columns=['stand_m_tov_nap'],
                         markers=['line'],
                         p=None,
                         plot_legend_names=[''],
                         plot_freq=[None],
                         tmin=None,
                         tmax=None,
                         hoover_names=['Peil'],
                         hoover_date_format="%Y-%m-%d",
                         ylabel='m NAP',
                         colors=['blue'],
                         add_filter_to_legend=False,
                         return_filename=False):
        """Create an interactive plot of the observations using bokeh.

        Todo:

        - add options for hoovers, markers, linestyle

        Parameters
        ----------
        savedir : str, optional
            directory used for the folium map and bokeh plots
        plot_columns : list of str, optional
            name of the column in the obs df that will be plotted with bokeh
        markers : list of str, optional
            type of markers that can be used for plot, 'line' and 'circle' are
            supported
        p : bokeh.plotting.figure, optional
            reference to existing figure, if p is None a new figure is created
        plot_legend_names : list of str, optional
            legend in bokeh plot
        plot_freq : list of str, optional
            bokeh plot is resampled with this frequency to reduce the size
        tmin : dt.datetime, optional
            start date for timeseries plot
        tmax : dt.datetime, optional
            end date for timeseries plot
        hoover_names : list of str, optional
            names will be displayed together with the plot_column value
            when hoovering over plot
        hoover_date_format : str, optional
            date format to use when hoovering over a plot
        ylabel : str, optional
            label on the y-axis
        colors : list of str, optional
            colors used for the plots
        add_filter_to_legend : boolean, optional
            if True the attributes bovenkant_filter and onderkant_filter
            are added to the legend name
        return_filename : boolean, optional
            if True filename will be returned

        Returns
        -------
        fname_plot : str or bokeh plot
            filename of the bokeh plot or reference to bokeh plot
        """

        from bokeh.plotting import figure
        from bokeh.models import ColumnDataSource, HoverTool
        from bokeh.plotting import save
        from bokeh.resources import CDN

        # create plot dataframe
        plot_df = self._obj[tmin:tmax].copy()
        plot_df['date'] = plot_df.index.strftime(hoover_date_format)
        if plot_df.empty or plot_df[plot_columns].isna().all().all():
            raise ValueError(
                '{} has no data between {} and {}'.format(self._obj.name, tmin, tmax))

        # create plot
        if p is None:
            p = figure(plot_width=600, plot_height=400, x_axis_type='datetime',
                       title='')
            p.yaxis.axis_label = ylabel

        # get x axis
        xcol = self._obj.index.name
        if xcol is None:
            xcol = 'index'

        # get color
        if len(colors) < len(plot_columns):
            colors = colors * len(plot_columns)

        # plot multiple columns
        for i, column in enumerate(plot_columns):
            # legend name
            if add_filter_to_legend:
                lname = '{} {} (NAP {:.2f} - {:.2f})'.format(plot_legend_names[i], self._obj.name,
                                                             self._obj.onderkant_filter,
                                                             self._obj.bovenkant_filter)
            else:
                lname = '{} {}'.format(plot_legend_names[i], self._obj.name)

            # resample data
            if plot_freq[i] is None:
                source = ColumnDataSource(plot_df[[column, 'date']])
            else:
                source = ColumnDataSource(
                    plot_df[[column, 'date']].resample(plot_freq[i]).nearest())

            # plot data
            if markers[i] == 'line':
                p.line(xcol, column, source=source, color=colors[i],
                       legend_label=lname,
                       alpha=0.8, muted_alpha=0.2)
            elif markers[i] == 'circle':
                p.circle(xcol, column, source=source, color=colors[i],
                         legend_label=lname,
                         alpha=0.8, muted_alpha=0.2)
            else:
                raise NotImplementedError("marker '{}' invalid. Only line and"
                                          "circle are currently available".format(markers[i]))

        # hoover options
        tooltips = []
        for i, column in enumerate(plot_columns):
            tooltips.append((hoover_names[i], "@{}".format(column)))
        tooltips.append(('date', "@date"))
        hover = HoverTool(tooltips=tooltips)

        p.add_tools(hover)

        p.legend.location = "top_left"
        p.legend.click_policy = "mute"

        # save plot
        if savedir is not None:
            if not os.path.isdir(savedir):
                os.makedirs(savedir)
            self._obj.iplot_fname = os.path.join(
                savedir, self._obj.name + '.html')
            save(p, self._obj.iplot_fname, resources=CDN, title=self._obj.name)

        if return_filename:
            return self._obj.iplot_fname
        else:
            return p



import os
import warnings

import numpy as np

from .. import util
from ..observation import ModelObs


def read_imod_results(obs_collection, ml, runfile, mtime, model_ws,
                      modelname='', nlay=None, exclude_layers=0,
                      verbose=False):
    """Read imod model results at point locations.

    Parameters
    ----------
    obs_collection : ObsCollection
        collection of observations at which points imod results will be read
    ml : flopy.modflow.mf.model
        modflow model
    runfile : Runfile
        imod runfile object
    mtime : list of datetimes
        datetimes corresponding to the model periods
    model_ws : str
        model workspace with imod model
    nlay : int, optional
        number of layers if None the number of layers from ml is used.
    modelname : str
        modelname
    exclude_layers : int
        exclude modellayers from being read from imod
    verbose : boolean, optional
        Print additional information to the screen (default is False).
    """
    import imod
    if ml.modelgrid.xoffset == 0 or ml.modelgrid.yoffset == 0:
        warnings.warn('you probably want to set the xll and/or yll '
                      'attributes of ml.modelgrid')

    if nlay is None:
        nlay = ml.modelgrid.nlay

    xmid, ymid, _ = ml.modelgrid.xyzcellcenters

    xy = np.array([xmid.ravel(), ymid.ravel()]).T
    uv = obs_collection.loc[:, ("x", "y")].dropna(how="any", axis=0).values
    vtx, wts = util.interp_weights(xy, uv)

    hm_ts = np.zeros((obs_collection.shape[0], len(mtime)))

    # loop over layers
    for m in range(nlay):
        if m < exclude_layers:
            continue
        mask = obs_collection.modellayer.values == m
        # loop over timesteps
        for t, date in enumerate(mtime):
            head_idf = 'head_{}_l{}.idf'.format(
                date.strftime('%Y%m%d'), m + 1)
            fname = os.path.join(
                model_ws,
                runfile.data['OUTPUTDIRECTORY'],
                'head',
                head_idf)
            if verbose:
                print('read {}'.format(fname))
            ihds, _attrs = imod.idf.read(fname)
            hm = util.interpolate(ihds, vtx, wts)
            hm_ts[mask, t] = hm[mask]

    mo_list = []
    for i, name in enumerate(obs_collection.index):
        mo = ModelObs(index=mtime,
                      data=hm_ts[i],
                      name=name,
                      model=modelname,
                      x=obs_collection.loc[name, 'x'],
                      y=obs_collection.loc[name, 'y'],
                      meta=obs_collection.loc[name, 'obs'].meta)
        mo_list.append(mo)

    return mo_list


def read_modflow_results(obs_collection, ml, hds_arr, mtime,
                         modelname='', nlay=None, exclude_layers=None,
                         verbose=False):
    """Read modflow groundwater heads at points in obs_collection.

    Parameters
    ----------
    obs_collection : ObsCollection
        locations of model observation
    ml : flopy.modflow.mf.model
        modflow model
    hds_arr : numpy array
        heads with shape (ntimesteps, nlayers, nrow, ncol)
    mtime : list of datetimes
        dates for each model timestep
    modelname : str, optional
        modelname
    nlay : int, optional
        number of layers if None the number of layers from ml is used.
    exclude_layers : list of int, optional
        exclude the observations in these modellayers
    verbose : boolean, optional
        Print additional information to the screen (default is False).
    """
    if ml.modelgrid.grid_type == 'structured':
        if ml.modelgrid.xoffset == 0 or ml.modelgrid.yoffset == 0:
            warnings.warn('you probably want to set the xll '
                          'and/or yll attributes in DIS!')

    if nlay is None:
        nlay = ml.modelgrid.nlay

    if modelname == '':
        modelname = ml.name

    xmid, ymid, _ = ml.modelgrid.xyzcellcenters

    if ml.modelgrid.grid_type == 'structured':
        xy = np.array([xmid.ravel(), ymid.ravel()]).T
    elif ml.modelgrid.grid_type == 'vertex':
        xy = np.array([xmid, ymid]).T

    uv = obs_collection.loc[:, ("x", "y")].dropna(how="any", axis=0).values
    vtx, wts = util.interp_weights(xy, uv)

    # get interpolated timeseries from hds_arr
    hm_ts = np.nan * np.ones((obs_collection.shape[0], hds_arr.shape[0]))

    if exclude_layers is None:
        exclude_layers = []

    # loop over layers
    for m in range(nlay):
        if m in exclude_layers:
            continue
        mask = obs_collection["modellayer"].values == m
        # loop over timesteps
        for t in range(hds_arr.shape[0]):
            ihds = hds_arr[t, m]
            ihds[ihds <= -999.] = np.nan
            hm = util.interpolate(ihds, vtx, wts)
            hm_ts[mask, t] = hm[mask]

    mo_list = []
    for i, name in enumerate(obs_collection.index):
        mo = ModelObs(index=mtime,
                      data=hm_ts[i],
                      name=name,
                      model=modelname,
                      x=obs_collection.loc[name, 'x'],
                      y=obs_collection.loc[name, 'y'],
                      meta=obs_collection.loc[name, 'obs'].meta)
        mo_list.append(mo)

    return mo_list



from timeit import default_timer

import arctic
from tqdm import tqdm


def read_arctic(connstr, libname, ObsClass, verbose=False):
    """Read all timeseries from Arctic database.

    Parameters
    ----------
    connstr : str
        connection string to MongoDB instance
    libname : str
        name of library to read data from
    ObsClass : Obs
        type of Obs to load data as
    verbose : bool, optional
        progressbar and timing info, by default False

    Returns
    -------
    obs_list : list of Obs
        list of Obs
    """
    arc = arctic.Arctic(connstr)
    lib = arc.get_library(libname)

    start = default_timer()
    obs_list = []
    rows_read = 0
    for sym in (tqdm(lib.list_symbols())
                if verbose else lib.list_symbols()):
        item = lib.read(sym)
        o = ObsClass(item.data, name=sym, meta=item.metadata)
        obs_list.append(o)
        if verbose:
            rows_read += len(item.data.index)
    end = default_timer()
    if verbose:
        print("Symbols: {0:.0f}  "
              "Rows: {1:,.0f}  "
              "Time: {2:.2f}s  "
              "Rows/s: {3:,.1f}".format(len(lib.list_symbols()),
                                        rows_read,
                                        (end - start),
                                        rows_read / (end - start)))

    return obs_list



import numpy as np
import pandas as pd

from . import accessor


def get_model_layer_z(z, zvec, left=-999, right=999,
                      verbose=False):
    """Get index of model layer based on elevation.

    Assumptions:

    - the highest value in zvec is the top of model layer 0.
    - if z is equal to the bottom of a layer, the model layer above that
      layer is assigned.

    Parameters
    ----------
    z : int, float
        elevation.
    zvec : list, np.array
        elevations of model layers. shape is nlay + 1
    left : int, optional
        if z is below the lowest value in zvec, this value is returned. 
        The default is -999.
    right : TYPE, optional
        if z is above the highest value in zvec, this value is returned. 
        The default is 999.

    Returns
    -------
    int : int
        model layer

    Examples
    --------
    >>> zvec = [0, -10, -20, -30]
    >>> get_model_layer_z(-5, zvec)
    0

    >>> get_model_layer_z(-25, zvec)
    2

    >>> get_model_layer_z(-50, zvec)
    -999

    >>> get_model_layer_z(100, zvec)
    999

    >>> get_model_layer_z(-20, zvec)
    1
    """
    # make sure zvec is descending
    zvec = np.sort(zvec)[::-1]
    if z in zvec:
        z += 1e-10
    lay = int(np.interp(z, zvec[::-1], np.arange(len(zvec))[::-1],
                        left=left, right=right))

    return lay


def check_if_var_is_invalid(var):

    if var is None:
        return True
    elif np.isnan(var):
        return True

    return False


def get_modellayer_from_filter(ftop, fbot, zvec, left=-999, right=999,
                               verbose=False):
    """


    Parameters
    ----------
    ftop : int or float
        top of filter.
    fbot : int or float
        bottom of filter, has to be lower than ftop.
    zvec : list or np.array
        elevations of the modellayers at the location of the filter.
    left : int, optional
        value to return if filter is below the modellayers. 
        The default is -999.
    right : int, optional
        value to return if filter is above the modellayers. 
        The default is-999.
    verbose : bool, optional
        DESCRIPTION. The default is False.

    Raises
    ------
    ValueError
        raised if something unexpected happens.

    Returns
    -------
    int or np.nan
        modellayer.

    examples
    --------
    >>> zvec = [0, -10, -20, -30, -40]
    >>> get_modellayer_from_filter(-5, -7, zvec)
    0

    >>> get_modellayer_from_filter(-25, -27, zvec)
    2

    >>> get_modellayer_from_filter(-15, -27, zvec, verbose=True)
    2

    >>> get_modellayer_from_filter(-5, -27, zvec, verbose=True)
    1

    >>> get_modellayer_from_filter(-5, -37, zvec, verbose=True)
    1

    >>> get_modellayer_from_filter(15, -5, zvec, verbose=True)
    0

    >>> get_modellayer_from_filter(15, 5, zvec, verbose=True)
    999

    >>> get_modellayer_from_filter(-55, -65, zvec, verbose=True)
    -999

    >>> get_modellayer_from_filter(15, -65, zvec, verbose=True)
    nan

    >>> get_modellayer_from_filter(None, -7, zvec, verbose=True)
    0

    >>> get_modellayer_from_filter(None, None, zvec, verbose=True)
    nan

    """

    zvec = np.sort(zvec)[::-1]
    ftop_invalid = check_if_var_is_invalid(ftop)
    fbot_invalid = check_if_var_is_invalid(fbot)
    if ftop_invalid and fbot_invalid:
        if verbose:
            print("- values for ftop and fbot are invalid!")
        return np.nan
    elif fbot_invalid:
        lay_ftop = get_model_layer_z(ftop, zvec,
                                     left=left, right=right)
        if verbose:
            print(f"- fbot invalid. selected based on ftop: {lay_ftop}")
        return lay_ftop
    elif ftop_invalid:
        lay_fbot = get_model_layer_z(fbot, zvec,
                                     left=left, right=right)
        if verbose:
            print(f"- ftop invalid. selected based on fbot: {lay_fbot}")
        return lay_fbot

    if ftop < fbot:
        if verbose:
            print('- filter top below filter bot, switch top and bot')
        fbot, ftop = ftop, fbot

    lay_ftop = get_model_layer_z(ftop, zvec,
                                 left=left, right=right)
    lay_fbot = get_model_layer_z(fbot, zvec,
                                 left=left, right=right)

    # Piezometer in layer in which majority of screen is located
    if lay_fbot == lay_ftop:
        if verbose:
            print(f"- selected layer {lay_fbot}")
        return lay_fbot

    else:
        if lay_fbot == left and lay_ftop == right:
            if verbose:
                print('- filter spans all layers. '
                      'return nan')
            return np.nan
        elif lay_ftop == right:
            if verbose:
                print('- filter top higher than top layer. '
                      'selected layer {}'.format(lay_fbot))
            return lay_fbot

        elif lay_fbot == left:
            if verbose:
                print('- filter bot lower than bottom layer. '
                      'selected layer {}'.format(lay_ftop))
            return lay_ftop

        if verbose:
            print("- filter crosses layer boundary:")
            print("  - layers: {0}, {1}".format(lay_ftop, lay_fbot))

        if verbose:
            print(f"- piezometer filter spans {lay_fbot - lay_ftop +1} layers."
                  " checking length per layer")
            print("  - length per layer:")

        # check which layer has the biggest length of the filter
        length_layers = np.zeros(lay_fbot - lay_ftop + 1)
        for i in range(len(length_layers)):
            if i == 0:
                length_layers[i] = ftop - zvec[lay_ftop + 1]
            elif (i + 1) == len(length_layers):
                length_layers[i] = zvec[lay_ftop + i] - fbot
            else:
                length_layers[i] = zvec[lay_ftop + i] - zvec[lay_ftop + 1 + i]

            if verbose:
                print(f"    - lay {lay_ftop+i}: {length_layers[i]:.2f}")

        # choose layer with biggest length
        rel_layer = np.argmax(length_layers)
        lay_out = lay_ftop + rel_layer
        if verbose:
            print(f"  - selected layer: {lay_out}")
        return lay_out

    return np.nan


def get_zvec(x, y, gwf=None):
    """get a list with the vertical layer boundaries at a point in the model.

    Parameters
    ----------
    x : int or float
        x coördinate.
    y : int or float
        y coordinate.
    gwf : flopy.mf6.modflow.mfgwf.ModflowGwf
        modflow model with top and bottoms

    Raises
    ------
    NotImplementedError
        not all grid types are supported yet.

    Returns
    -------
    zvec : list
        list of vertical layer boundaries. length is nlay + 1.
    """

    if gwf is not None:
        if gwf.modelgrid.grid_type == 'structured':
            r, c = gwf.modelgrid.intersect(x, y, forgive=True)
            if np.isfinite(r) & np.isfinite(c):
                zvec = [gwf.modelgrid.top[r, c]] + [gwf.modelgrid.botm[i, r, c]
                                                    for i in range(gwf.modelgrid.nlay)]
            else:
                print("Point is not within model extent! Returning NaN.")
                zvec = np.nan
        elif gwf.modelgrid.grid_type == 'vertex':
            idx = gwf.modelgrid.intersect(x, y, forgive=True)
            if np.isfinite(idx):
                zvec = [gwf.modelgrid.top[idx]] + [gwf.modelgrid.botm[i, idx]
                                                   for i in range(gwf.modelgrid.nlay)]
            else:
                print("Point is not within model extent! Returning NaN.")
                zvec = np.nan
        else:
            raise NotImplementedError(
                f'gridtype {gwf.modelgrid.grid_type} not (yet) implemented')

    return zvec


@accessor.register_obscollection_accessor("gwobs")
class GwObsAccessor:

    def __init__(self, oc_obj):
        self._obj = oc_obj

    def set_filter_num(self, radius=1, xcol='x', ycol='y', if_exists='error',
                       add_to_meta=False):
        """This method computes the filternumbers based on the location of the
        observations.

        Then it sets the value of the filternumber:

        - in the ObsCollection dataframe
        - as the attribute of an Obs object
        - in the meta dictionary of the Obs object (only if add_to_meta is
          True)


        This method is useful for groundwater observations. If two or more
        observation points are close to each other they will be seen as one
        location with multiple filters. The filternr is based on the
        'onderkant_filter' attribute of the observations in such a way that
        the deepest filter has the highest filter number.

        Parameters
        ----------
        radius : int, optional
            max distance between two observations to be seen as one location, by default 1
        xcol : str, optional
            column name with x coordinates, by default 'x'
        ycol : str, optional
            column name with y coordinates, by default 'y'
        if_exists : str, optional
            what to do if an observation point already has a filternr, options:
            'error', 'replace' or 'keep', by default 'error'
        add_to_meta : bool, optional
            if True the filternr is added to the meta dictionary
            of an observation. The default is False.

        Raises
        ------
        RuntimeError
            if the column filternr exists and if_exists='error' an error is
            raised
        """

        # check if column exists in obscollection
        if 'filternr' in self._obj.columns:
            # set type to numeric
            if self._obj['filternr'].dtype != np.number:
                self._obj['filternr'] = pd.to_numeric(
                    self._obj['filternr'], errors='coerce')

            # check if name should be replaced
            if if_exists == 'error':
                raise RuntimeError(
                    "the column 'filternr' already exist, set if_exists='replace' to replace the current values")
            elif if_exists == 'replace':
                self._obj['filternr'] = np.nan
        else:
            self._obj['filternr'] = np.nan

        # ken filternummers toe aan peilbuizen die dicht bij elkaar staan
        for name in self._obj.index:
            if np.isnan(self._obj.loc[name, 'filternr']):
                x = self._obj.loc[name, xcol]
                y = self._obj.loc[name, ycol]
                distance_to_other_filters = np.sqrt(
                    (self._obj[xcol] - x)**2 + (self._obj[ycol] - y)**2)
                dup_x = self._obj.loc[distance_to_other_filters < radius]
                if dup_x.shape[0] == 1:
                    self._obj._set_metadata_value(name, 'filternr', 1,
                                                  add_to_meta=add_to_meta)
                else:
                    dup_x2 = dup_x.sort_values(
                        'onderkant_filter', ascending=False)
                    for i, pb_dub in enumerate(dup_x2.index):
                        self._obj._set_metadata_value(pb_dub, 'filternr', i + 1,
                                                      add_to_meta=add_to_meta)

    def set_filter_num_location(self, loc_col, radius=1, xcol='x', ycol='y',
                                if_exists='error', add_to_meta=False):
        """This method sets the filternr and locatie name of an observation
        point based on the location of the observations.

        When two or more filters are close to another, as defined by radius,
        they are set to the same `locatie` and an increasing `filternr` based
        on depth.

        The value of the filternumber and the location are set:

        - in the ObsCollection dataframe
        - as the attribute of an Obs object
        - in the meta dictionary of the Obs object (only if add_to_meta is
          True)

        This method is useful for groundwater observations. If two or more
        observation points are close to each other they will be seen as one
        location with multiple filters. The filternr is based on the
        'onderkant_filter' attribute of the observations in such a way that
        the deepest filter has the highest filter number. The locatie is
        based on the named of the loc_col of the filter with the lowest
        filternr.

        Parameters
        ----------
        loc_col : str
            the column name with the names to use for the locatie
        radius : int, optional
            max distance between two observations to be seen as one location, by default 1
        xcol : str, optional
            column name with x coordinates, by default 'x'
        ycol : str, optional
            column name with y coordinates, by default 'y'
        if_exists : str, optional
            what to do if an observation point already has a filternr, options:
            'error', 'replace' or 'keep', by default 'error'
        add_to_meta : bool, optional
            if True the filternr and location are added to the meta dictionary
            of an observation. The default is False.

        Raises
        ------
        RuntimeError
            if the column filternr exists and if_exists='error' an error
            is raised
        """

        # check if columns exists in obscollection
        if 'filternr' in self._obj.columns or 'locatie' in self._obj.columns:
            if if_exists == 'error':
                raise RuntimeError(
                    "the column 'filternr or locatie' already exist, set if_exists='replace' to replace the current values")
            elif if_exists == 'replace':
                self._obj['filternr'] = np.nan
                self._obj['locatie'] = np.nan
        else:
            self._obj['filternr'] = np.nan
            self._obj['locatie'] = np.nan

        # ken filternummers toe aan peilbuizen die dicht bij elkaar staan
        for name in self._obj.index:
            if np.isnan(self._obj.loc[name, 'filternr']):
                x = self._obj.loc[name, xcol]
                y = self._obj.loc[name, ycol]
                distance_to_other_filters = np.sqrt(
                    (self._obj[xcol] - x)**2 + (self._obj[ycol] - y)**2)
                dup_x = self._obj.loc[distance_to_other_filters < radius]
                if dup_x.shape[0] == 1:
                    self._obj._set_metadata_value(name, 'filternr', 1,
                                                  add_to_meta=add_to_meta)
                    locatie = self._obj.loc[name, loc_col]
                    self._obj._set_metadata_value(name, 'locatie', locatie,
                                                  add_to_meta=add_to_meta)
                else:
                    dup_x2 = dup_x.sort_values(
                        'onderkant_filter', ascending=False)
                    for i, pb_dub in enumerate(dup_x2.index):
                        if i == 0:
                            locatie = self._obj.loc[pb_dub, loc_col]
                        self._obj._set_metadata_value(pb_dub, 'filternr', i + 1,
                                                      add_to_meta=add_to_meta)
                        self._obj._set_metadata_value(pb_dub, 'locatie', locatie,
                                                      add_to_meta=add_to_meta)

    def get_modellayers(self, gwf, verbose=False):
        """Get the modellayer per observation. The layers can be obtained from
        the modflow model or can be defined in zgr.

        Parameters
        ----------
        gwf : flopy.mf6.modflow.mfgwf.ModflowGwf
            modflow model
        verbose : boolean, optional
            Print additional information to the screen (default is False).

        Returns
        -------
        pd.Series with the modellayers of each observation
        """
        modellayers = []
        for o in self._obj.obs.values:
            if verbose:
                print("-" * 10 + "\n " + o.name + ":")

            modellayers.append(o.gwobs.get_modellayer_modflow(gwf,
                                                              verbose=verbose))

        modellayers = pd.Series(index=self._obj.index, data=modellayers,
                                name='modellayer')

        return modellayers

    def get_regis_layers(self, verbose=False):
        """Get the regis layer per observation.

        Parameters
        ----------
        verbose : boolean, optional
            Print additional information to the screen (default is False).

        Returns
        -------
        pd.Series with the names of the regis layer of each observation
        """

        return self._obj.obs.apply(lambda o: o.gwobs.get_regis_layer(verbose))


@accessor.register_obs_accessor("gwobs")
class GeoAccessorObs:
    def __init__(self, obs):
        self._obj = obs

    def get_modellayer_modflow(self, gwf, left=-999, right=999, verbose=False):
        """Add modellayer to meta dictionary.

        Parameters
        ----------
        gwf : flopy.mf6.modflow.mfgwf.ModflowGwf
            modflow model
        verbose : boolean, optional
            Print additional information to the screen (default is False).

        Returns
        -------
        int
            modellayer
        """

        zvec = get_zvec(self._obj.x, self._obj.y, gwf)

        if np.all(np.isnan(zvec)):
            return np.nan
        else:
            modellayer = get_modellayer_from_filter(self._obj.bovenkant_filter,
                                                    self._obj.onderkant_filter,
                                                    zvec,
                                                    left=left, right=right,
                                                    verbose=verbose)
            return modellayer

    def get_regis_layer(self, verbose=False):
        """find the name of the REGIS layer based on the filter depth.

        Parameters
        ----------
        verbose : boolean, optional
            Print additional information to the screen (default is False).

        Returns
        -------
        str
            name of REGIS layer
        """
        import xarray as xr

        if np.isnan(self._obj.bovenkant_filter) or np.isnan(self._obj.onderkant_filter):
            return 'nan'

        # connect to regis netcdf
        regis_url = r'http://www.dinodata.nl:80/opendap/REGIS/REGIS.nc'
        regis_ds = xr.open_dataset(regis_url, decode_times=False)

        # rename layer in regis netcdf
        regis_ds = regis_ds.rename({'layer': 'layer_old'})
        regis_ds.coords['layer'] = regis_ds.layer_old.astype(str)
        regis_ds = regis_ds.swap_dims({'layer_old': 'layer'})

        # get zvec regis netcdf
        z = regis_ds['bottom'].sel(x=self._obj.x, y=self._obj.y,
                                   method='nearest').dropna(dim='layer')
        ztop = regis_ds['top'].sel(x=self._obj.x, y=self._obj.y,
                                   method='nearest').max()
        zvec = np.concatenate([[ztop.values], z.values])

        # get index of regis model layer
        layer_i = get_modellayer_from_filter(self._obj.bovenkant_filter,
                                             self._obj.onderkant_filter,
                                             zvec,
                                             left=-999, right=999,
                                             verbose=verbose)

        if layer_i == 999:
            return 'above'
        elif layer_i == -999:
            return 'below'

        return str(z.layer[layer_i].values)



import os

import numpy as np
import pandas as pd

from ..util import get_files


def _read_wiski_header(f, header_sep=":", header_identifier='#',
                       end_header_str=None):
    line = f.readline()
    header = dict()
    while header_identifier in line:
        prop, val = line.split(header_sep)
        prop = prop.strip().replace(header_identifier, "")
        val = val.strip()
        try:
            val = float(val)
        except ValueError:
            pass
        header[prop] = val
        line = f.readline()
        if end_header_str is not None:
            if end_header_str in line:
                break

    return line, header


def read_wiski_file(fname, sep=";", header_sep=None, header_identifier='#',
                    read_series=True, infer_datetime_format=True,
                    translate_dic={}, verbose=False,
                    tz_localize=True, to_mnap=True, **kwargs):
    if verbose:
        print('reading -> {}'.format(os.path.split(fname)[-1]))

    # manually break header parse at certain point
    if "end_header_str" in kwargs.keys():
        end_header_str = kwargs.pop("end_header_str")
    else:
        end_header_str = None

    # read header
    with open(fname, "r") as f:
        if header_sep is None:
            line, header = _read_wiski_header(f, end_header_str=end_header_str,
                                              header_identifier=header_identifier)
        else:
            line, header = _read_wiski_header(f, header_sep=header_sep,
                                              header_identifier=header_identifier,
                                              end_header_str=end_header_str)

        # specify names
        for key, item in translate_dic.items():
            header[key] = header[item]

        # get column names of data
        # columns = split(r'\s{2,}', line)
        if sep == r"\s+":
            columns = line.split("\t")
        else:
            columns = line.split(sep)
        columns = [icol.strip("\n") for icol in columns]
        columns = [icol.replace("[", "_[") for icol in columns]

        validator = np.lib._iotools.NameValidator()
        columns = validator(columns)

        if read_series:
            # read data
            data = pd.read_csv(f, sep=sep, header=None, names=columns,
                               infer_datetime_format=infer_datetime_format,
                               **kwargs)

            if tz_localize:
                data.index = data.index.tz_localize(None)

            # convert Value to float
            col = [icol for icol in data.columns if
                   icol.lower().startswith("value")][0]

            data[col] = pd.to_numeric(
                data[col], errors="coerce")

            if to_mnap:
                data.rename(columns={col: 'stand_m_tov_nap'}, inplace=True)
        else:
            data = None

        # translate some header keys
        metadata = {}
        for key, val in header.items():
            if key == 'Station Site':
                metadata['locatie'] = val
            elif key == 'x':
                metadata['x'] = val
            elif key == "y":
                metadata['y'] = val
            elif key == 'name':
                metadata['name'] = val
            # this adds the other header keys to metadata
            # but this will not work if keys contain spaces etc.
            # because it tries adding them as attributes.
            # else:
            #     metadata[key] = val

    return data, metadata


def read_wiski_dir(dirname, ObsClass=None, suffix=".csv",
                   unpackdir=None, force_unpack=False, preserve_datetime=False,
                   keep_all_obs=True, verbose=True, **kwargs):

    # get files
    dirname, unzip_fnames = get_files(dirname, ext=suffix,
                                      force_unpack=force_unpack,
                                      preserve_datetime=preserve_datetime)

    if not unzip_fnames:
        raise FileNotFoundError("no files were found in '{}' that end with '{}'".format(
            os.path.join(dirname), suffix))

    # gather all obs in list
    obs_list = []
    for i, csv in enumerate(unzip_fnames):
        if verbose:
            print("reading {0}/{1} -> {2}".format(i +
                                                  1, len(unzip_fnames), csv))
        obs = ObsClass.from_wiski(os.path.join(
            dirname, csv), verbose=False, **kwargs)

        if obs.metadata_available:
            obs_list.append(obs)
        elif keep_all_obs:
            obs_list.append(obs)
        else:
            if verbose:
                print('not added to collection -> {}'.format(csv))

    return obs_list



"""module with ObsCollection class for a collection of observations.

The ObsCollection class is a subclass of a pandas DataFrame with
additional attributes and methods.

More information about subclassing pandas DataFrames can be found here:
http://pandas.pydata.org/pandas-docs/stable/development/extending.html#extending-subclassing-pandas
"""
import os
import warnings

import numpy as np
import pandas as pd

from . import observation as obs
from . import util


class ObsCollection(pd.DataFrame):
    """class for a collection of point observations.

    An ObsCollection object is a subclass of a pandas.DataFrame and allows for
    additional attributes and methods. Additional attributes are
    defined in the '_metadata' attribute.

    Parameters
    ----------
    name : str
        name of the observation collection
    meta : dic
        metadata of the observation collection
    """
    # temporary properties
    _internal_names = pd.DataFrame._internal_names + ['none']
    _internal_names_set = set(_internal_names)

    # normal properties
    _metadata = ['name',
                 'meta',
                 ]

    def __init__(self, *args, **kwargs):
        """constructor of the ObsCollection.

        *args must be input for the pandas.DataFrame constructor,
        **kwargs can be one of the attributes listed in _metadata or
        keyword arguments for the constructor of a pandas.DataFrame.
        """
        self.name = kwargs.pop('name', '')
        self.meta = kwargs.pop('meta', {})

        super(ObsCollection, self).__init__(*args, **kwargs)

    @property
    def _constructor(self):
        return ObsCollection

    def _infer_otype(self, verbose=False):
        """Infer observation type from the obs column.

        Parameters
        ----------
        verbose : boolean, optional
            Print additional information to the screen (default is False).

        Returns
        -------
        otype, otypes
            type or list of types of the observation objects
        """
        otypes = self.obs.apply(lambda x: type(x)).unique()
        if otypes.shape[0] == 1:
            if verbose:
                print('inferred observation type: {}'.format(otypes[0]))
            return otypes[0]
        elif otypes.shape[0] > 1:
            if verbose:
                print('inferred multiple otypes, types: {}'.format(otypes))
            return otypes
        else:
            raise TypeError('could not infer observation type')

    def _set_metadata_value(self, iname, att_name, value, add_to_meta=False,
                            verbose=False):
        """ Set a value on three different levels at once:
            1. the value in an ObsCollection DataFrame
            2. the attribute of the observation
            3. the value in the meta dictionary of an observation (optional)

        Parameters
        ----------
        iname : str, int, float, ...
            observation name. Must be same type as self.index.
            e.g. B52D0111_3
        att_name : str, int, float, ...
            name of the column in self.columns and attribute
            of the observation. e.g. 'x'
        value : str, int, float, ...
            value of the the att_name. e.g. 116234
        add_to_meta : bool, optional
            if True the att_name, value pair is added to the meta dictionary
            of an observation. The default is False.
        verbose : boolean, optional
            Print additional information to the screen (default is False).

        Raises
        ------
        ValueError
            if the iname is not in self.index the value cannot be set.

        Returns
        -------
        None.

        """
        if iname not in self.index:
            raise ValueError(f"{iname}  not in index")

        self.loc[iname, att_name] = value
        if verbose:
            print(f'set {iname}, {att_name} to {value}')

        o = self.loc[iname, 'obs']
        if att_name in o._metadata:
            setattr(o, att_name, value)
            if verbose:
                print(f'set attribute {att_name} of {iname} to {value}')

        if add_to_meta:
            o.meta.update({att_name: value})
            if verbose:
                print(f'add {att_name} of {iname} with value {value} to meta')

    @classmethod
    def from_arctic(cls, connstr, libname, ObsClass=obs.GroundwaterObs,
                    verbose=False):
        """Load ObsCollection from MongoDB using arctic.

        Parameters
        ----------
        connstr : str
            database connection string
        libname : str
            library name
        ObsClass : class, optional
            observation class to store single timeseries, by
            default obs.GroundwaterObs
        verbose : bool, optional
            show progress bar and database read summary, by default False

        Returns
        -------
        ObsCollection
            ObsCollection DataFrame containing all the obs
        """
        from .io.io_arctic import read_arctic

        meta = {'type': obs.GroundwaterObs}
        obs_list = read_arctic(connstr, libname, ObsClass, verbose=verbose)
        obs_df = util._obslist_to_frame(obs_list)

        return cls(obs_df, meta=meta)

    @classmethod
    def from_dataframe(cls, df, obs_list=None, ObsClass=obs.GroundwaterObs):
        """Create an observation collection from a DataFrame by adding a column
        with empty observations.

        Parameters
        ----------
        df : pandas DataFrame
            input dataframe. If this dataframe has a column named 'obs' the
            column is replaced with empty observation objects.
        obs_list : list of observation.Obs, optional
            list of observations. Default is None
        ObsClass : class, optional
            observation class used to create empty obs object, by
            default obs.GroundwaterObs

        Returns
        -------
        ObsCollection
            ObsCollection DataFrame with the 'obs' column
        """
        meta = {'type': obs.GroundwaterObs}
        if isinstance(df, pd.DataFrame):
            if obs_list is None:
                obs_list = [ObsClass() for i in range(len(df))]
            df['obs'] = obs_list
        else:
            raise TypeError(
                f'df should be type pandas.DataFrame not {type(df)}')

        return cls(df, meta=meta)

    @classmethod
    def from_dino(cls,
                  dirname=None,
                  extent=None,
                  bbox=None,
                  locations=None,
                  ObsClass=obs.GroundwaterObs,
                  subdir='Grondwaterstanden_Put',
                  suffix='1.csv',
                  unpackdir=None,
                  force_unpack=False,
                  preserve_datetime=False,
                  keep_all_obs=True,
                  name=None,
                  verbose=False,
                  **kwargs):
        """Read dino data within an extent from the server or from a directory
        with downloaded files.

        Parameters
        ----------
        dirname : str, optional
            directory name, can be a .zip file or the parent directory
            of subdir
        extent : list, tuple or numpy-array (user must specify extent or bbox)
            get dinodata online within this extent [xmin, xmax, ymin, ymax]
        bbox : list, tuple or numpy-array (user must specify extent or bbox)
            The bounding box, in RD-coordinates, for which you want to
            retrieve locations [xmin, ymin, xmax, ymax]
        locations : list of str, optional
            list of names with location and filter number, separated by
            'filtersep'
        ObsClass : type
            class of the observations, so far only GroundwaterObs is supported
        subdir : str
            subdirectory of dirname with data files
        suffix : str
            suffix of files in subdir that will be read
        unpackdir : str
            destination directory of the unzipped file
        force_unpack : boolean, optional
            force unpack if dst already exists
        preserve_datetime : boolean, optional
            use date of the zipfile for the destination file
        keep_all_obs : boolean, optional
            add all observation points to the collection, even the points
            without measurements or metadata
        name : str, optional
            the name of the observation collection
        verbose : boolean, optional
            Print additional information to the screen (default is False).
        kwargs:
            kwargs are passed to the io_dino.download_dino_within_extent() or
            the io_dino.read_dino_dir() function

        Returns
        -------
        cls(obs_df) : ObsCollection
            collection of multiple point observations
        """
        from .io.io_dino import (read_dino_dir, download_dino_within_extent,
                                 download_dino_groundwater_bulk)

        if dirname is not None:
            # read dino directory
            if name is None:
                name = subdir

            meta = {'dirname': dirname,
                    'type': ObsClass,
                    'suffix': suffix,
                    'unpackdir': unpackdir,
                    'force_unpack': force_unpack,
                    'preserve_datetime': preserve_datetime,
                    'verbose': verbose,
                    'keep_all_obs': keep_all_obs
                    }

            obs_list = read_dino_dir(dirname,
                                     ObsClass,
                                     subdir,
                                     suffix,
                                     unpackdir,
                                     force_unpack,
                                     preserve_datetime,
                                     verbose,
                                     keep_all_obs,
                                     **kwargs)

        elif extent is not None or bbox is not None:
            # read dino data within extent
            if ObsClass == obs.GroundwaterObs:
                layer = 'grondwatermonitoring'
            else:
                raise NotImplementedError(
                    'cannot download {} from Dino'.format(ObsClass))

            if name is None:
                name = '{} from DINO'.format(layer)

            meta = kwargs.copy()
            meta.update({'verbose': verbose,
                         'extent': extent,
                         'bbox': bbox,
                         'layer': layer,
                         'keep_all_obs': keep_all_obs,
                         'verbose': verbose})

            obs_list = download_dino_within_extent(
                extent=extent, bbox=bbox, ObsClass=ObsClass, layer=layer,
                keep_all_obs=keep_all_obs, verbose=verbose, **kwargs)

        elif locations is not None:
            name = "DINO"

            meta = {'dirname': dirname,
                    'type': ObsClass,
                    'verbose': verbose}

            obs_list = download_dino_groundwater_bulk(locations,
                                                      ObsClass=ObsClass,
                                                      verbose=verbose,
                                                      **kwargs)
        else:
            raise ValueError("No data source provided!")

        obs_df = util._obslist_to_frame(obs_list)
        return cls(obs_df, name=name, meta=meta)

    @classmethod
    def from_dino_server(cls, extent=None, bbox=None,
                         ObsClass=obs.GroundwaterObs,
                         name=None, keep_all_obs=True,
                         verbose=False, **kwargs
                         ):
        """Read dino data from a server.

        Parameters
        ----------
        extent : list, tuple or numpy-array (user must specify extent or bbox)
            The extent, in RD-coordinates, for which you want to retreive locations
            [xmin, xmax, ymin, ymax]
        bbox : list, tuple or numpy-array (user must specify extent or bbox)
            The bounding box, in RD-coordinates, for which you want to retreive locations
            [xmin, ymin, xmax, ymax]
        ObsClass : type
            class of the observations, so far only GroundwaterObs is supported
        name : str, optional
            the name of the observation collection
        keep_all_obs : boolean, optional
            add all observation points to the collection, even without data or
            metadata
        verbose : boolean, optional
            Print additional information to the screen (default is False).
        kwargs:
            kwargs are passed to the io_dino.download_dino_within_extent() function

        Returns
        -------
        cls(obs_df) : ObsCollection
            collection of multiple point observations
        """

        warnings.warn(
            "this method will be removed in future versions,"
            " use from_dino instead", DeprecationWarning)

        from .io.io_dino import download_dino_within_extent

        if ObsClass == obs.GroundwaterObs:
            layer = 'grondwatermonitoring'
        else:
            raise NotImplementedError(
                'cannot download {} from Dino'.format(ObsClass))

        if name is None:
            name = '{} from DINO'.format(layer)

        meta = kwargs.copy()
        meta.update({'verbose': verbose})

        obs_list = download_dino_within_extent(
            extent=extent, bbox=bbox, ObsClass=ObsClass, layer=layer,
            keep_all_obs=keep_all_obs, verbose=verbose, **kwargs)

        obs_df = util._obslist_to_frame(obs_list)

        if bbox is None:
            bbox = [extent[0], extent[2], extent[1], extent[3]]

        return cls(obs_df, name=name, meta=meta)

    @classmethod
    def from_dino_dir(cls,
                      dirname=None,
                      ObsClass=obs.GroundwaterObs,
                      subdir='Grondwaterstanden_Put',
                      suffix='1.csv',
                      unpackdir=None,
                      force_unpack=False,
                      preserve_datetime=False,
                      keep_all_obs=True,
                      name=None,
                      verbose=False,
                      **kwargs):
        """Read a dino directory.

        Parameters
        ----------
        dirname : str, optional
            directory name, can be a .zip file or the parent directory of subdir
        ObsClass : type
            class of the observations, e.g. GroundwaterObs or WaterlvlObs
        subdir : str
            subdirectory of dirname with data files
        suffix : str
            suffix of files in subdir that will be read
        unpackdir : str
            destination directory of the unzipped file
        force_unpack : boolean, optional
            force unpack if dst already exists
        preserve_datetime : boolean, optional
            use date of the zipfile for the destination file
        keep_all_obs : boolean, optional
            add all observation points to the collection, even without data or
            metadata
        name : str, optional
            the name of the observation collection
        verbose : boolean, optional
            Print additional information to the screen (default is False).
        kwargs:
            kwargs are passed to the io_dino.read_dino_dir() function

        Returns
        -------
        cls(obs_df) : ObsCollection
            collection of multiple point observations
        """

        warnings.warn(
            "this method will be removed in future "
            "versions, use from_dino instead", DeprecationWarning)

        from .io.io_dino import read_dino_dir

        if name is None:
            name = subdir

        meta = {'dirname': dirname,
                'type': ObsClass,
                'suffix': suffix,
                'unpackdir': unpackdir,
                'force_unpack': force_unpack,
                'preserve_datetime': preserve_datetime,
                'verbose': verbose,
                'keep_all_obs': keep_all_obs
                }

        obs_list = read_dino_dir(
            dirname,
            ObsClass,
            subdir,
            suffix,
            unpackdir,
            force_unpack,
            preserve_datetime,
            verbose,
            keep_all_obs,
            **kwargs)

        obs_df = util._obslist_to_frame(obs_list)

        return cls(obs_df, name=name, meta=meta)

    @classmethod
    def from_artdino_dir(
            cls,
            dirname=None,
            ObsClass=obs.GroundwaterObs,
            subdir='csv',
            suffix='.csv',
            unpackdir=None,
            force_unpack=False,
            preserve_datetime=False,
            keep_all_obs=True,
            name=None,
            verbose=False,
            **kwargs):
        """Read a dino directory.

        Parameters
        ----------
        extent : list, optional
            get dinodata online within this extent [xmin, xmax, ymin, ymax]
        dirname : str, optional
            directory name, can be a .zip file or the parent directory of subdir
        ObsClass : type
            class of the observations, e.g. GroundwaterObs or WaterlvlObs
        subdir : str
            subdirectory of dirname with data files
        suffix : str
            suffix of files in subdir that will be read
        unpackdir : str
            destination directory of the unzipped file
        force_unpack : boolean, optional
            force unpack if dst already exists
        preserve_datetime : boolean, optional
            use date of the zipfile for the destination file
        keep_all_obs : boolean, optional
            add all observation points to the collection, even without data or
            metadata
        name : str, optional
            the name of the observation collection
        verbose : boolean, optional
            Print additional information to the screen (default is False).
        kwargs:
            kwargs are passed to the io_dino.read_dino_dir() function

        Returns
        -------
        cls(obs_df) : ObsCollection
            collection of multiple point observations
        """

        from .io.io_dino import read_artdino_dir

        if name is None:
            name = subdir

        meta = {'dirname': dirname,
                'type': ObsClass,
                'suffix': suffix,
                'unpackdir': unpackdir,
                'force_unpack': force_unpack,
                'preserve_datetime': preserve_datetime,
                'verbose': verbose,
                'keep_all_obs': keep_all_obs
                }

        obs_list = read_artdino_dir(
            dirname,
            ObsClass,
            subdir,
            suffix,
            unpackdir,
            force_unpack,
            preserve_datetime,
            verbose,
            keep_all_obs,
            **kwargs)

        obs_df = util._obslist_to_frame(obs_list)

        return cls(obs_df, name=name, meta=meta)

    @classmethod
    def from_fews_xml(cls, file_or_dir=None,
                      xmlstring=None, ObsClass=obs.GroundwaterObs,
                      name='fews', translate_dic=None, filterdict=None,
                      locations=None, to_mnap=True, remove_nan=True,
                      low_memory=True, unpackdir=None, force_unpack=False,
                      preserve_datetime=False, verbose=False):
        """Read one or several FEWS PI-XML files.

        Parameters
        ----------
        file_or_dir :  str
            zip, xml or directory with zips or xml files to read

        ObsClass : type
            class of the observations, e.g. GroundwaterObs or WaterlvlObs
        name : str, optional
            name of the observation collection, 'fews' by default
        translate_dic : dic or None, optional
            translate names from fews. If None this default dictionary is used:
            {'locationId': 'locatie'}.
        filterdict : dict, optional
            dictionary with tag name to apply filter to as keys, and list of
            accepted names as dictionary values to keep in final result,
            i.e. {"locationId": ["B001", "B002"]}
        locations : list of str, optional
            list of locationId's to read from XML file, others are skipped.
            If None (default) all locations are read. Only supported by
            low_memory=True method!
        to_mnap : boolean, optional
            if True a column with 'stand_m_tov_nap' is added to the dataframe
        remove_nan : boolean, optional
            remove nan values from measurements, flag information about the
            nan values is also lost
        low_memory : bool, optional
            whether to use xml-parsing method with lower memory footprint,
            default is True
        unpackdir : str
            destination directory to unzip file if fname is a .zip
        force_unpack : boolean, optional
            force unpack if dst already exists
        preserve_datetime : boolean, optional
            whether to preserve datetime from zip archive
        verbose : boolean, optional
            Print additional information to the screen (default is False).

        Returns
        -------
        cls(obs_df) : ObsCollection
            collection of multiple point observations
        """
        from .io.io_fews import read_xml_filelist, read_xmlstring

        if translate_dic is None:
            translate_dic = {'locationId': 'locatie'}

        meta = {'type': ObsClass,
                'verbose': verbose}

        if (file_or_dir is not None):
            # get files
            dirname, unzip_fnames = util.get_files(
                file_or_dir, ext=".xml", unpackdir=unpackdir,
                force_unpack=force_unpack, preserve_datetime=preserve_datetime)
            meta.update({'filename': dirname})

            obs_list = read_xml_filelist(unzip_fnames,
                                         ObsClass,
                                         directory=dirname,
                                         translate_dic=translate_dic,
                                         filterdict=filterdict,
                                         locations=locations,
                                         to_mnap=to_mnap,
                                         remove_nan=remove_nan,
                                         low_memory=low_memory,
                                         verbose=verbose)

            obs_df = util._obslist_to_frame(obs_list)
            return cls(obs_df, name=name, meta=meta)

        elif (file_or_dir is None) and (xmlstring is not None):
            obs_list = read_xmlstring(xmlstring,
                                      ObsClass,
                                      translate_dic=translate_dic,
                                      filterdict=filterdict,
                                      locationIds=locations,
                                      low_memory=low_memory,
                                      to_mnap=to_mnap,
                                      remove_nan=remove_nan,
                                      verbose=verbose)
            obs_df = util._obslist_to_frame(obs_list)
            return cls(obs_df, name=name, meta=meta)

        else:
            raise ValueError(
                'either specify variables file_or_dir or xmlstring')

    @classmethod
    def from_fieldlogger(cls, fname, name='', ObsClass=obs.GroundwaterObs):
        """Read a fieldlogger file into a list of observation objects.

        Parameters
        ----------
        fname : str
            name of the fieldlogger location csv file
        name : str, optional
            name of the observation collection
        ObsClass : observation class
            type of fieldlogger observations
        """

        from .io.io_fieldlogger import fieldlogger_csv_to_obs_list

        obs_list, fieldlogger_meta = fieldlogger_csv_to_obs_list(
            fname, ObsClass=obs.GroundwaterObs)
        obs_df = util._obslist_to_frame(obs_list)

        return cls(obs_df, meta=fieldlogger_meta)

    @classmethod
    def from_imod(cls, obs_collection, ml, runfile, mtime, model_ws,
                  modelname='', nlay=None, exclude_layers=0, verbose=False):
        """Read imod model results at point locations.

        Parameters
        ----------
        obs_collection : ObsCollection
            collection of observations at which points imod results will be read
        ml : flopy.modflow.mf.model
            modflow model
        runfile : Runfile
            imod runfile object
        mtime : list of datetimes
            datetimes corresponding to the model periods
        model_ws : str
            model workspace with imod model
        nlay : int, optional
            number of layers if None the number of layers from ml is used.
        modelname : str
            modelname
        exclude_layers : int
            exclude modellayers from being read from imod
        verbose : boolean, optional
            Print additional information to the screen (default is False).
        """
        from .io.io_modflow import read_imod_results
        mo_list = read_imod_results(obs_collection, ml, runfile,
                                    mtime, model_ws,
                                    modelname=modelname, nlay=nlay,
                                    exclude_layers=exclude_layers,
                                    verbose=verbose)
        obs_df = util._obslist_to_frame(mo_list)
        return cls(obs_df, name=modelname)

    @classmethod
    def from_knmi(cls, locations=None, stns=None, xmid=None, ymid=None,
                  meteo_vars=["RD"], name='', start=[None, None],
                  end=[None, None], ObsClass=obs.KnmiObs, **kwargs):
        """get knmi observations from a list of locations or a list of
        stations.

        Parameters
        ----------
        locations : pd.DataFrame or None
            dataframe with x and y coordinates. The default is None
        stns : list of str or None
            list of knmi stations. The default is None
        xmid : np.array, optional
            x coördinates of the cell centers of your grid shape(ncol)
        ymid : np.array, optional
            y coördinates of the cell centers of your grid shape(nrow)
        meteo_vars : list or tuple of str
            meteo variables e.g. ["RD", "EV24"]. The default is ("RD").
            See list of all possible variables below
        name : str, optional
            name of the obscollection. The default is ''.
        start : list of str, datetime or None]
            start date of observations per meteo variable. The default is
            [None, None]
        end : list of str, datetime or None]
            end date of observations per meteo variable. The default is
            [None, None]
        ObsClass : type or None
            class of the observations, only KnmiObs is supported for now. The
            default is None
        **kwargs :
            kwargs are passed to the io_knmi.get_knmi_obslist function

        List of possible meteo variables:
            DDVEC     = Vectorgemiddelde windrichting in graden (360=noord, 90=oost, 180=zuid, 270=west, 0=windstil/variabel). Zie http://www.knmi.nl/kennis-en-datacentrum/achtergrond/klimatologische-brochures-en-boeken / Vector mean wind direction in degrees (360=north, 90=east, 180=south, 270=west, 0=calm/variable)
            FHVEC     = Vectorgemiddelde windsnelheid (in 0.1 m/s). Zie http://www.knmi.nl/kennis-en-datacentrum/achtergrond/klimatologische-brochures-en-boeken / Vector mean windspeed (in 0.1 m/s)
            FG        = Etmaalgemiddelde windsnelheid (in 0.1 m/s) / Daily mean windspeed (in 0.1 m/s)
            FHX       = Hoogste uurgemiddelde windsnelheid (in 0.1 m/s) / Maximum hourly mean windspeed (in 0.1 m/s)
            FHXH      = Uurvak waarin FHX is gemeten / Hourly division in which FHX was measured
            FHN       = Laagste uurgemiddelde windsnelheid (in 0.1 m/s) / Minimum hourly mean windspeed (in 0.1 m/s)
            FHNH      = Uurvak waarin FHN is gemeten / Hourly division in which FHN was measured
            FXX       = Hoogste windstoot (in 0.1 m/s) / Maximum wind gust (in 0.1 m/s)
            FXXH      = Uurvak waarin FXX is gemeten / Hourly division in which FXX was measured
            TG        = Etmaalgemiddelde temperatuur (in 0.1 graden Celsius) / Daily mean temperature in (0.1 degrees Celsius)
            TN        = Minimum temperatuur (in 0.1 graden Celsius) / Minimum temperature (in 0.1 degrees Celsius)
            TNH       = Uurvak waarin TN is gemeten / Hourly division in which TN was measured
            TX        = Maximum temperatuur (in 0.1 graden Celsius) / Maximum temperature (in 0.1 degrees Celsius)
            TXH       = Uurvak waarin TX is gemeten / Hourly division in which TX was measured
            T10N      = Minimum temperatuur op 10 cm hoogte (in 0.1 graden Celsius) / Minimum temperature at 10 cm above surface (in 0.1 degrees Celsius)
            T10NH     = 6-uurs tijdvak waarin T10N is gemeten / 6-hourly division in which T10N was measured; 6=0-6 UT, 12=6-12 UT, 18=12-18 UT, 24=18-24 UT
            SQ        = Zonneschijnduur (in 0.1 uur) berekend uit de globale straling (-1 voor <0.05 uur) / Sunshine duration (in 0.1 hour) calculated from global radiation (-1 for <0.05 hour)
            SP        = Percentage van de langst mogelijke zonneschijnduur / Percentage of maximum potential sunshine duration
            Q         = Globale straling (in J/cm2) / Global radiation (in J/cm2)
            DR        = Duur van de neerslag (in 0.1 uur) / Precipitation duration (in 0.1 hour)
            RH        = Etmaalsom van de neerslag (in 0.1 mm) (-1 voor <0.05 mm) / Daily precipitation amount (in 0.1 mm) (-1 for <0.05 mm)
            RHX       = Hoogste uursom van de neerslag (in 0.1 mm) (-1 voor <0.05 mm) / Maximum hourly precipitation amount (in 0.1 mm) (-1 for <0.05 mm)
            RHXH      = Uurvak waarin RHX is gemeten / Hourly division in which RHX was measured
            PG        = Etmaalgemiddelde luchtdruk herleid tot zeeniveau (in 0.1 hPa) berekend uit 24 uurwaarden / Daily mean sea level pressure (in 0.1 hPa) calculated from 24 hourly values
            PX        = Hoogste uurwaarde van de luchtdruk herleid tot zeeniveau (in 0.1 hPa) / Maximum hourly sea level pressure (in 0.1 hPa)
            PXH       = Uurvak waarin PX is gemeten / Hourly division in which PX was measured
            PN        = Laagste uurwaarde van de luchtdruk herleid tot zeeniveau (in 0.1 hPa) / Minimum hourly sea level pressure (in 0.1 hPa)
            PNH       = Uurvak waarin PN is gemeten / Hourly division in which PN was measured
            VVN       = Minimum opgetreden zicht / Minimum visibility; 0: <100 m, 1:100-200 m, 2:200-300 m,..., 49:4900-5000 m, 50:5-6 km, 56:6-7 km, 57:7-8 km,..., 79:29-30 km, 80:30-35 km, 81:35-40 km,..., 89: >70 km)
            VVNH      = Uurvak waarin VVN is gemeten / Hourly division in which VVN was measured
            VVX       = Maximum opgetreden zicht / Maximum visibility; 0: <100 m, 1:100-200 m, 2:200-300 m,..., 49:4900-5000 m, 50:5-6 km, 56:6-7 km, 57:7-8 km,..., 79:29-30 km, 80:30-35 km, 81:35-40 km,..., 89: >70 km)
            VVXH      = Uurvak waarin VVX is gemeten / Hourly division in which VVX was measured
            NG        = Etmaalgemiddelde bewolking (bedekkingsgraad van de bovenlucht in achtsten, 9=bovenlucht onzichtbaar) / Mean daily cloud cover (in octants, 9=sky invisible)
            UG        = Etmaalgemiddelde relatieve vochtigheid (in procenten) / Daily mean relative atmospheric humidity (in percents)
            UX        = Maximale relatieve vochtigheid (in procenten) / Maximum relative atmospheric humidity (in percents)
            UXH       = Uurvak waarin UX is gemeten / Hourly division in which UX was measured
            UN        = Minimale relatieve vochtigheid (in procenten) / Minimum relative atmospheric humidity (in percents)
            UNH       = Uurvak waarin UN is gemeten / Hourly division in which UN was measured
            EV24      = Referentiegewasverdamping (Makkink) (in 0.1 mm) / Potential evapotranspiration (Makkink) (in 0.1 mm)
        """

        from .io.io_knmi import get_knmi_obslist

        meta = {}
        meta['start'] = start
        meta['end'] = end
        meta['name'] = name
        meta['ObsClass'] = ObsClass
        meta['meteo_vars'] = meteo_vars

        obs_list = get_knmi_obslist(locations, stns,
                                    xmid, ymid,
                                    meteo_vars,
                                    ObsClass=ObsClass,
                                    start=start,
                                    end=end,
                                    **kwargs)

        obs_df = util._obslist_to_frame(obs_list)

        return cls(obs_df, name=name, meta=meta)

    @classmethod
    def from_list(cls, obs_list, name='', verbose=False):
        """read observations from a list of obs objects.

        Parameters
        ----------
        obs_list : list of observation.Obs
            list of observations
        name : str, optional
            name of the observation collection
        verbose : boolean, optional
            Print additional information to the screen (default is False).
        """
        obs_df = util._obslist_to_frame(obs_list)
        return cls(obs_df, name=name)

    @classmethod
    def from_menyanthes(cls, fname, name='', ObsClass=obs.GroundwaterObs,
                        verbose=False):

        from .io.io_menyanthes import read_file

        menyanthes_meta = {'filename': fname,
                           'type': ObsClass,
                           'verbose': verbose}

        obs_list = read_file(fname, ObsClass, verbose)
        obs_df = util._obslist_to_frame(obs_list)

        return cls(obs_df, meta=menyanthes_meta)

    @classmethod
    def from_modflow(cls, obs_collection, ml, hds_arr, mtime,
                     modelname='', nlay=None, exclude_layers=None, verbose=False):
        """Read modflow groundwater heads at points in obs_collection.

        Parameters
        ----------
        obs_collection : ObsCollection
            locations of model observation
        ml : flopy.modflow.mf.model
            modflow model
        hds_arr : numpy array
            heads with shape (ntimesteps, nlayers, nrow, ncol)
        mtime : list of datetimes
            dates for each model timestep
        modelname : str, optional
            modelname
        nlay : int, optional
            number of layers if None the number of layers from ml is used.
        exclude_layers : list of int, optional
            exclude the observations in these model layers
        verbose : boolean, optional
            Print additional information to the screen (default is False).
        """
        from .io.io_modflow import read_modflow_results
        mo_list = read_modflow_results(obs_collection, ml, hds_arr,
                                       mtime, modelname=modelname,
                                       nlay=nlay,
                                       exclude_layers=exclude_layers,
                                       verbose=verbose)
        obs_df = util._obslist_to_frame(mo_list)

        return cls(obs_df)

    @classmethod
    def from_pastas_project(cls, pr, ObsClass=obs.GroundwaterObs,
                            name=None, pr_meta=None, rename_dic={}):

        from .io.io_pastas import read_project

        if pr_meta is None:
            pr_meta = pr.file_info

        if name is None:
            name = pr.name

        obs_list = read_project(pr, obs.GroundwaterObs,
                                rename_dic=rename_dic)
        obs_df = util._obslist_to_frame(obs_list)
        return cls(obs_df, meta=pr_meta, name=name)

    @classmethod
    def from_pystore(cls, storename, pystore_path,
                     ObsClass=obs.GroundwaterObs,
                     extent=None, collection_names=None,
                     item_names=None, nameby="item",
                     read_series=True, verbose=True, progressbar=False):
        """Create ObsCollection from pystore store.

        Parameters
        ----------
        storename : str
            name of the store
        pystore_path : str
            path in which stores are stored
        ObsClass : type of Obs, optional
            the Observation type used for reading in the individual series,
            by default obs.GroundwaterObs
        extent : list or tuple, optional
            if not None only Observations within this extent are read
            [xmin, xmax, ymin, ymax]
        collection_names : list of str
            collection names that will be extracted from the store. Default is
            None which reads all collections.
        item_names : list of str
            item (Observation) names that will be extracted from the store.
            The other items (Observations) will be ignored. if None all items
            are read.
        read_series : bool, optional
            if False, read only metadata, default is True which
            loads the full dataset
        nameby : str
            pick whether Obs are named by 'item' or 'collection'
        progressbar : bool, optional
            whether to show progress bar, default is False, for
            best result set verbose to False!

        Returns
        -------
        ObsCollection
            Collection of observations
        """

        from .io.io_pystore import read_pystore

        meta = {'fname': os.path.join(pystore_path, storename),
                'verbose': verbose}

        obs_list = read_pystore(storename, pystore_path,
                                obs.GroundwaterObs,
                                extent=extent,
                                collection_names=collection_names,
                                item_names=item_names, nameby=nameby,
                                read_series=read_series, verbose=verbose,
                                progressbar=progressbar)
        # if read series is False, returns dataframe with only metadata
        if isinstance(obs_list, list):
            obs_df = util._obslist_to_frame(obs_list)
            meta['type'] = obs.GroundwaterObs
        else:
            # only metadata in df
            obs_df = obs_list

        return cls(obs_df, name=storename, meta=meta)

    @classmethod
    def from_waterinfo(cls, file_or_dir, name="", ObsClass=obs.WaterlvlObs,
                       progressbar=True):
        """Read waterinfo file or directory.

        Parameters
        ----------
        file_or_dir : str
            path to file or directory. Files can be .csv or .zip
        name : str, optional
            name of the collection, by default ""
        ObsClass : Obs, optional
            type of Obs to read data as, by default obs.WaterlvlObs
        progressbar : bool, optional
            show progressbar, by default True

        Returns
        -------
        ObsCollection
            ObsCollection containing data
        """
        from .io import io_waterinfo

        meta = {
            'name': name,
            'type': ObsClass,
            'filename': file_or_dir
        }

        obs_list = io_waterinfo.read_waterinfo_obs(
            file_or_dir, ObsClass, progressbar=progressbar)
        obs_df = util._obslist_to_frame(obs_list)

        return cls(obs_df, name=name, meta=meta)

    @classmethod
    def from_wiski(cls, dirname, ObsClass=obs.GroundwaterObs, suffix='.csv',
                   unpackdir=None, force_unpack=False, preserve_datetime=False,
                   verbose=False, keep_all_obs=True, **kwargs):

        from .io.io_wiski import read_wiski_dir

        meta = {'dirname': dirname,
                'type': ObsClass,
                'suffix': suffix,
                'unpackdir': unpackdir,
                'force_unpack': force_unpack,
                'preserver_datetime': preserve_datetime,
                'verbose': verbose,
                'keep_all_obs': keep_all_obs
                }

        name = "wiski_import"
        obs_list = read_wiski_dir(dirname,
                                  ObsClass=ObsClass,
                                  suffix=suffix,
                                  unpackdir=unpackdir,
                                  force_unpack=force_unpack,
                                  preserve_datetime=preserve_datetime,
                                  verbose=verbose,
                                  keep_all_obs=keep_all_obs,
                                  **kwargs)
        obs_df = util._obslist_to_frame(obs_list)

        return cls(obs_df, name=name, meta=meta)

    def to_fieldlogger(self, fname, additional_collection=None, otype=None,
                       infer_otype=True,
                       use_default_otype_settings=True,
                       heading=None,
                       name=None, subname=None, inputfield=None,
                       properties=None, group=None, group_color='blue',
                       verbose=False):
        """Write a csv file that can be read by fieldlogger.

        Notes
        -----
        1. option to change input fields for measurements is not yet added.
        2. option to add min/max values for measurements is not yet added.

        Parameters
        ----------
        fname : str
            name of the fieldlogger csv file
        additional_collections : list, optional
            additional ObsCollections that are added to the fieldlogger csv
        otype : str, optional
            observation type, 'groundwater' and 'surfacewater' are supported
        infer_otype : boolean, optional
            infer mtype from the obs column in self.obs, default is True
        use_default_otype_settings : boolean, optional
            use the default settings for this measurement type
        heading : str, optional
            heading of the csv, if None use default heading
        name : series or str, optional
            1st column in fieldlogger csv, if None use index of ObsCollection
        subname : series or str, optional
            2nd column in fieldlogger csv, if None use name + _sub
        inputfield : series or str, optional
            3th column in fieldlogger csv, if None use 'Stand|Opmerking'
        properties : series or str, optional
            4th column in fieldlogger csv, if None use empty string
        group : series or str, optional
            5th column in fieldlogger csv, if None do not add group to csv
        group_color : series or str, optional
            color of the group
        """
        from .io import io_fieldlogger

        if infer_otype:
            otype = self._infer_otype(verbose)
            if not isinstance(
                    otype, type):  # check of dit ook echt een type is
                raise TypeError(
                    'could not infer observation type, use infer_otype=False')

        f_df = io_fieldlogger.df_to_fieldlogger_csv(
            self,
            fname,
            otype,
            use_default_otype_settings,
            heading,
            name,
            subname,
            inputfield,
            properties,
            group,
            group_color,
            verbose)

        return f_df

    def to_pi_xml(self, fname, timezone="", version="1.24"):
        from .io import io_fews
        io_fews.write_pi_xml(self, fname, timezone=timezone, version=version)

    def to_pystore(self, store_name, pystore_path, groupby, item_name=None,
                   overwrite=False):
        """Write timeseries and metadata to Pystore format. Series are grouped
        by 'groupby'. Each group is a Collection, each series within that group
        an Item.

        Parameters
        ----------
        store_name : str
            name of the store
        pystore_path : str
            path where store should be saved
        groupby : str
            column name to groupby, (for example, group by location,
            which would create a collection per location, and write
            all timeseries at that location as Items into that Collection).
        item_name : str, optional
            name of column to use as item name, default is None, Item then
            takes name from obs.name
        overwrite : bool, optional
            if True overwrite current data in store, by default False
        """
        import pystore
        pystore.set_path(pystore_path)
        store = pystore.store(store_name)
        for name, group in self.groupby(by=groupby):
            # Access a collection (create it if not exist)
            collection = store.collection(name, overwrite=overwrite)
            for i, o in enumerate(group.obs):
                imeta = o.meta.copy()
                if 'datastore' in imeta.keys():
                    imeta['datastore'] = str(imeta['datastore'])
                # add extra columns to item metadata
                for icol in group.columns:
                    if icol != "obs" and icol != 'meta':
                        # check if type is numpy integer
                        # numpy integers are not json serializable
                        if isinstance(group.iloc[i].loc[icol], np.integer):
                            imeta[icol] = int(group.iloc[i].loc[icol])
                        else:
                            imeta[icol] = group.iloc[i].loc[icol]
                if item_name is None:
                    name = o.name
                else:
                    name = o.meta[item_name]
                collection.write(name, o, metadata=imeta,
                                 overwrite=overwrite)

    def to_arctic(self, connstr, libname, verbose=False):
        """Write ObsCollection to MongoDB using Arctic.

        Parameters
        ----------
        connstr : str
            database connection string
        libname : str
            name of the library to store data
        verbose : bool, optional
            show progress bar, by default False
        """
        import arctic
        from tqdm import tqdm
        arc = arctic.Arctic(connstr)

        # get library
        try:
            lib = arc.get_library(libname)
        except arctic.exceptions.LibraryNotFoundException:
            print("Library '{}' not found, initializing library...".format(
                libname))
            arc.initialize_library(libname)
            lib = arc[libname]

        # write data
        for o in (tqdm(self.obs) if verbose else self.obs):
            metadata = o.meta
            lib.write(o.name, o, metadata=metadata)

    def to_gdf(self, xcol='x', ycol='y'):
        """convert ObsCollection to GeoDataFrame.

        Parameters
        ----------
        xcol : str
            column name with x values
        ycol : str
            column name with y values

        Returns
        -------
        gdf : geopandas.GeoDataFrame
        """
        return util.df2gdf(self, xcol, ycol)

    def to_report_table(self, columns=['locatie', 'filternr',
                                       'Van', 'Tot', '# metingen']):

        if 'Van' in columns:
            self['Van'] = self.obs.apply(lambda x: x.index[0])
        if 'Tot' in columns:
            self['Tot'] = self.obs.apply(lambda x: x.index[-1])
        if '# metingen' in columns:
            self['# metingen'] = self.obs.apply(lambda x: x.shape[0])

        return self[columns]

    def to_pastastore(self, pstore=None, pstore_name='',
                      obs_column='stand_m_tov_nap',
                      kind='oseries', add_metadata=True,
                      overwrite=False, verbose=False):
        """add observations to a new or existing pastastore.

        Parameters
        ----------
        oc : observation.ObsCollection
            collection of observations
        pstore : pastastore.PastasProject, optional
            Existing pastastore, if None a new project is created
        pstore_name : str, optional
            Name of the pastastore only used if pstore is None
        obs_column : str, optional
            Name of the column in the Obs dataframe to be used
        kind : str, optional
            The kind of series that is added to the pastas project
        add_metadata : boolean, optional
            If True metadata from the observations added to the project
        overwrite : boolean, optional
            if True, overwrite existing series in pastastore, default is False
        verbose : boolean, optional
            Print additional information to the screen (default is False).

        Returns
        -------
        project : pastastore.PastasProject
            the pastas project with the series from the ObsCollection
        """
        from .io.io_pastas import create_pastastore

        pstore = create_pastastore(self, pstore, pstore_name,
                                   add_metadata=add_metadata,
                                   kind=kind,
                                   obs_column=obs_column,
                                   verbose=verbose, overwrite=overwrite)

        return pstore

    def to_pastas_project(self, pr=None, project_name='',
                          obs_column='stand_m_tov_nap',
                          kind='oseries', add_metadata=True,
                          verbose=False, **kwargs):
        """add observations to a new or existing pastas project.

        Parameters
        ----------
        pr : pastas.project, optional
            Existing pastas project, if None a new project is created
        project_name : str, optional
            Name of the pastas project only used if pr is None
        obs_column : str, optional
            Name of the column in the Obs dataframe to be used
        kind : str, optional
            The kind of series that is added to the pastas project
        add_metadata : boolean, optional
            If True metadata from the observations added to the project.
        verbose : boolean, optional
            Print additional information to the screen (default is False).
        kwargs
            arguments are passed to the create_pastas_project funtion

        Returns
        -------
        pr : pastas.project
            the pastas project with the series from the ObsCollection
        """
        warnings.warn(
            "this method will be removed in future versions, use to_pastastore instead", DeprecationWarning)

        from .io.io_pastas import create_pastas_project

        project = create_pastas_project(self, pr=pr,
                                        project_name=project_name,
                                        obs_column=obs_column,
                                        kind=kind, add_metadata=add_metadata,
                                        verbose=verbose, **kwargs)

        return project

    def to_shapefile(self, fname, xcol='x', ycol='y'):
        """save ObsCollection as shapefile.

        Parameters
        ----------
        fname : str
            filename of shapefile, ends with .shp
        xcol : str
            column name with x values
        ycol : str
            column name with y values
        """
        gdf = util.df2gdf(self, xcol, ycol)

        # remove obs column
        if 'obs' in gdf.columns:
            gdf.drop(columns='obs', inplace=True)

        # cast boolean columns to int
        for colname, coltype in gdf.dtypes.items():
            if coltype == bool:
                gdf[colname] = gdf[colname].astype(int)
        gdf.to_file(fname)

    def add_meta_to_df(self, key):
        """Get the values from the meta dictionary of each observation object
        and add these to the ObsCollection as a column.

        to the ObsCollection

        Parameters
        ----------
        key : str
            key in meta dictionary of observation object
        """

        self[key] = [o.meta[key] for o in self.obs.values]

    def get_series(self, tmin=None, tmax=None, col="stand_m_tov_nap"):
        if tmin is None:
            tmin = self.dates_first_obs.min()
        if tmax is None:
            tmax = self.dates_last_obs.max()
        return self.obs.apply(lambda o: o.loc[tmin:tmax, col])
